__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, grfstrings, nmlop
from nml.actions import action4, action6, action7, actionD, base_action
from nml.actions.action0properties import Action0Property, BaseAction0Property, properties, two_byte_property
from nml.ast import general


class BlockAllocation:
    """
    Administration of allocated blocks of a size, in a range of addresses.
    Blocks always start at address C{0}, but the first available freely usable
    address may be further.

    The allocation information is kept in L{allocated}. It has the following cases
    - An address does not exist in the dictionary.
      The address is free for use.
    - An address exists, with an integer number as size.
      At this address a block has been allocated with that size.
    - An address exists, with value C{None}.
      The address is part of an allocated block, but not the start.

    @ivar first: First freely usable address.
    @type first: C{int}

    @ivar last: Last freely usable address.
    @type last: C{int}

    @ivar name: Name for debug output.
    @type name: C{str}

    @ivar dynamic_allocation: True, if ids are allocated. False, if they refer to static entities.
    @type dynamic_allocation: C{bool}

    @ivar allocated: Mapping of allocated blocks.
    @type allocated: C{dict} of C{int} to allocation information.

    @ivar filled: Mapping of block size to the smallest address that may contain free space.
                  Serves as a cache to speed up searches.
    @type filled: C{dict} of C{int}
    """

    def __init__(self, first, last, name, dynamic_allocation=True):
        self.first = first
        self.last = last
        self.name = name
        self.dynamic_allocation = dynamic_allocation
        self.allocated = {}
        self.filled = {}

    def get_num_allocated(self):
        """
        Return number of allocated ids.
        """
        if self.dynamic_allocation:
            return len(self.allocated)
        else:
            return 0

    def get_max_allocated(self):
        """
        Return maximum number of allocatable ids.
        """
        if self.dynamic_allocation:
            return self.last - self.first + 1
        else:
            return 0

    def in_range(self, addr, length):
        """
        Does a block starting at the provided address with the given length in
        the available address space?

        @param addr: First address of the block.
        @type  addr: C{int}

        @param length: Number of addresses in the block.
        @type  length: C{int}

        @return: Whether the block fits entirely in the available address space.
        @rtype:  C{bool}
        """
        return addr >= 0 and addr + length - 1 <= self.last

    def get_size(self, addr):
        """
        Get the size of the block allocated at the given address.

        @param addr: Address being queried.
        @type  addr: C{int}

        @return: Size of the allocated block, if one was allocated here, else C{None}.
        @rtype:  C{int} or C{None}
        """
        return self.allocated.get(addr)

    def is_address_free(self, addr):
        """
        Is the space at the given address still available?
        ( cheaper variant of C{self.get_last_used(addr, 1) is None} )

        @param addr: Address being queried.
        @type  addr: C{int}

        @return: Whether the space at the provided address is available.
        @rtype:  C{bool}
        """
        return addr not in self.allocated

    def get_last_used(self, addr, length):
        """
        Check whether a block of addresses is used.

        @param addr: First address of the block.
        @type  addr: C{int}

        @param length: Number of addresses in the block.
        @type  length: C{int}

        @return: The last used address in the block, or C{None} if all addresses are free.
        @rtype:  C{int} or C{None}

        @precond: Addresses of the range should be within the available address space.
        """
        for idx in range(addr + length - 1, addr - 1, -1):
            if idx in self.allocated:
                return idx
        return None

    def mark_used(self, addr, length):
        """
        Mark an area as being allocated.

        @param addr: First address of the block.
        @type  addr: C{int}

        @param length: Number of addresses in the block.
        @type  length: C{int}

        @precond: Addresses of the block should be within the freely available address space.
        @precond: No address in the block may have been allocated.
        """
        self.allocated[addr] = length
        for idx in range(addr + 1, addr + length):
            self.allocated[idx] = None

    def find_unused(self, length):
        """
        Find an area of unused space.

        @param length: Number of addresses to find.
        @type  length: C{int}

        @return: Address in the freely available address space for the new block,
                 or C{None} if no space is available.
        @rtype:  C{int} or C{None}
        """
        idx = self.filled.get(length)
        if idx is None:
            # Never searched before with this block size.
            # Start at the biggest offset previously discovered with a smaller block size.
            smaller_filleds = [min_f for sz, min_f in self.filled.items() if sz < length]
            idx = self.first if len(smaller_filleds) == 0 else max(smaller_filleds)

        last_idx = self.last - length + 1
        while idx < last_idx:
            last_used = self.get_last_used(idx, length)
            if last_used is None:
                self.filled[length] = idx + length
                return idx

            idx = last_used + 1

        return None


# Available IDs for each feature.
# Maximum allowed id (houses and indtiles in principle allow up to 511, but action3 does not accept extended bytes).
used_ids = [
    BlockAllocation(116, 0xFFFF, "Train"),
    BlockAllocation(88, 0xFFFF, "Road Vehicle"),
    BlockAllocation(11, 0xFFFF, "Ship"),
    BlockAllocation(41, 0xFFFF, "Aircraft"),
    BlockAllocation(0, 0xFFFE, "Station"),  # UINT16_MAX - 1
    BlockAllocation(0, 8, "Canal", False),
    BlockAllocation(0, 15, "Bridge", False),
    BlockAllocation(0, 4095, "House"),
    BlockAllocation(0, -1, "Global", False),
    BlockAllocation(0, 254, "Industry Tile"),
    BlockAllocation(0, 127, "Industry"),
    BlockAllocation(0, 63, "Cargo"),
    BlockAllocation(0, -1, "Sound"),
    BlockAllocation(0, 127, "Airport"),
    BlockAllocation(0, -1, "Signal", False),
    BlockAllocation(0, 64000, "Object"),
    BlockAllocation(0, 63, "Railtype"),
    BlockAllocation(0, 255, "Airport Tile"),
    BlockAllocation(0, 62, "Roadtype"),
    BlockAllocation(0, 62, "Tramtype"),
    BlockAllocation(0, 0xFFFE, "RoadStop"),  # UINT16_MAX - 1
]


def print_stats():
    """
    Print statistics about used ids.
    """
    for feature in used_ids:
        used = feature.get_num_allocated()
        if used > 0 and feature.dynamic_allocation:
            generic.print_info("{} items: {}/{}".format(feature.name, used, feature.get_max_allocated()))


def mark_id_used(feature, id, num_ids):
    """
    Mark a range of ids as used.

    @param feature: Feature of the ids.
    @type  feature: C{int}

    @param id: First id to be marked.
    @type  id: C{int}

    @param num_ids: Number of ids to mark as used, starting with \a id.
    @type  num_ids: C{int}
    """
    used_ids[feature].mark_used(id, num_ids)


def check_id_range(feature, id, num_ids, pos):
    """
    Check that the id is valid, and is either free or points to an already allocated block.

    @param feature: Feature of the ids.
    @type  feature: C{int}

    @param id: Base id number.
    @type  id: C{int}

    @param num_ids: Number of ids to test, starting with \a id.
    @type  num_ids: C{int}

    @param pos: Source position of the check, for reporting errors.
    @type  pos: L{Position}

    @return: Whether the block was allocated already.
    @rtype:  C{bool}
    """
    blk_alloc = used_ids[feature]

    # Check that IDs are valid and in range.
    if not blk_alloc.in_range(id, num_ids):
        msg = "Item ID must be in range 0..{:d}, encountered {:d}..{:d}."
        msg = msg.format(blk_alloc.last, id, id + num_ids - 1)
        raise generic.ScriptError(msg, pos)

    # ID already defined, but with the same size: OK
    if blk_alloc.get_size(id) == num_ids:
        return True

    # All IDs free: no problem.
    if blk_alloc.get_last_used(id, num_ids) is None:
        return False

    # No space at the indicated position, report an error.

    if blk_alloc.get_size(id) is not None:
        # ID already defined with a different size: error.
        raise generic.ScriptError(
            "Item with ID {:d} has already been defined, but with a different size.".format(id), pos
        )

    if blk_alloc.is_address_free(id):
        # First item id free -> any of the additional tile ids must be blocked.
        msg = "This multi-tile house requires that item IDs {:d}..{:d} are free, but they are not."
        msg = msg.format(id, id + num_ids - 1)
        raise generic.ScriptError(msg, pos)

    # ID already defined as part of a multi-tile house.
    raise generic.ScriptError("Item ID {:d} has already used as part of a multi-tile house.".format(id), pos)


def get_free_id(feature, num_ids, pos):
    """
    Find an id to allocate a range of \a num_ids ids in a feature.

    @param feature: Feature of the ids.
    @type  feature: C{int}

    @param num_ids: Number of ids to allocate.
    @type  num_ids: C{int}

    @param pos: Position information.
    @type  pos: L{Position}
    """
    blk_alloc = used_ids[feature]

    addr = blk_alloc.find_unused(num_ids)
    if addr is None:
        msg = "Unable to allocate ID for item, no more free IDs available (maximum is {:d})"
        msg = msg.format(blk_alloc.last)
        raise generic.ScriptError(msg, pos)

    blk_alloc.mark_used(addr, num_ids)
    return addr


# Number of tiles for various house sizes
house_sizes = {
    0: 1,  # 1x1
    2: 2,  # 2x1
    3: 2,  # 1x2
    4: 4,  # 2x2
}


def adjust_value(value, org_value, unit, ottd_convert_func):
    """
    Make sure that the property value written to the NewGRF will match exactly
    the value as quoted

    @param value: The value to check, converted to base units
    @type value: L{Expression}

    @param org_value: The original value as written in the input file
    @type org_value: L{Expression}

    @param unit: The unit of the org_value
    @type unit: L{Unit} or C{None}

    @return: The adjusted value
    @rtype: L{Expression}
    """
    while ottd_convert_func(value, unit) > org_value.value:
        value = expression.ConstantNumeric(int(value.value - 1), value.pos)
    lower_value = value
    while ottd_convert_func(value, unit) < org_value.value:
        value = expression.ConstantNumeric(int(value.value + 1), value.pos)
    higher_value = value

    lower_value_ottd = abs(ottd_convert_func(lower_value, unit) - org_value.value)
    higher_value_ottd = abs(ottd_convert_func(higher_value, unit) - org_value.value)
    if lower_value_ottd < higher_value_ottd:
        return lower_value
    return higher_value


class Action0(base_action.BaseAction):
    """
    Representation of NFO action 0. Action 0 is used to set properties on all
    kinds of objects. It can set any number of properties on a list of
    consecutive IDs. Each property is defined by a unique (per feature) integer.

    @ivar feature: Feature number to set properties for.
    @type feature: C{int}

    @ivar id: First ID to set properties for
    @type id: C{int}

    @ivar prop_list: List of all properties that are to be set.
    @type prop_list: C{list} of L{BaseAction0Property}

    @ivar num_ids: Number of IDs to set properties for.
    @type num_ids: C{int} or C{None}
    """

    def __init__(self, feature, id):
        self.feature = feature
        self.id = id
        self.prop_list = []
        self.num_ids = None

    def prepare_output(self, sprite_num):
        if self.num_ids is None:
            self.num_ids = 1

    def write(self, file):
        size = 7
        for prop in self.prop_list:
            assert isinstance(prop, BaseAction0Property), type(prop)
            if isinstance(prop, Action0Property):
                assert len(prop.values) == self.num_ids
            size += prop.get_size()
        file.start_sprite(size)
        file.print_bytex(0)
        file.print_bytex(self.feature)
        file.print_byte(len(self.prop_list))
        file.print_bytex(self.num_ids)
        file.print_bytex(0xFF)
        file.print_wordx(self.id)
        file.newline()
        for prop in self.prop_list:
            prop.write(file)
        file.end_sprite()


def create_action0(feature, id, act6, action_list):
    """
    Create an action0 with variable id

    @param feature: Feature of the action0
    @type feature: C{int}

    @param id: ID of the corresponding item
    @type id: L{Expression}

    @param act6: Action6 to add any modifications to
    @type act6: L{Action6}

    @param action_list: Action list to append any extra actions to
    @type action_list: C{list} of L{BaseAction}

    @return: A tuple of (resulting action0, offset to use for action6)
    @rtype: C{tuple} of (L{Action0}, C{int})
    """
    id, offset = actionD.write_action_value(id, action_list, act6, 5, 2)

    action0 = Action0(feature, id.value)
    return (action0, offset)


def get_property_info_list(feature, name):
    """
    Find information on a single property, based on feature and name/number

    @param feature: Feature of the associated item
    @type feature: C{int}

    @param name: Name (or number) of the property
    @type name: L{Identifier} or L{ConstantNumeric}

    @return: A list of dictionaries with property information
    @rtype: C{list} of C{dict}
    """
    # Validate feature
    assert feature in range(0, len(properties))  # guaranteed by item
    if properties[feature] is None:
        raise generic.ScriptError(
            "Setting properties for feature '{}' is not possible, no properties are defined.".format(
                general.feature_name(feature)
            ),
            name.pos,
        )

    if isinstance(name, expression.Identifier):
        prop_name = name.value
        if prop_name not in properties[feature]:
            raise generic.ScriptError("Unknown property name: " + prop_name, name.pos)
        prop_info_list = properties[feature][prop_name]
        if not isinstance(prop_info_list, list):
            prop_info_list = [prop_info_list]
    elif isinstance(name, expression.ConstantNumeric):
        for prop_info_list in properties[feature].values():
            if not isinstance(prop_info_list, list):
                prop_info_list = [prop_info_list]
            # Only non-compound properties may be set by number
            if len(prop_info_list) == 1 and "num" in prop_info_list[0] and prop_info_list[0]["num"] == name.value:
                break
        else:
            raise generic.ScriptError("Unknown property number: " + str(name), name.pos)
    else:
        raise AssertionError()

    for prop_info in prop_info_list:
        if "replaced_by" in prop_info:
            generic.print_warning(
                generic.Warning.DEPRECATION,
                "'{}' is deprecated, consider using '{}' instead".format(prop_name, prop_info["replaced_by"]),
                name.pos,
            )
        if "warning" in prop_info:
            generic.print_warning(generic.Warning.GENERIC, prop_info["warning"], name.pos)
    return prop_info_list


def parse_property_value(prop_info, value, unit=None, size_bit=None):
    """
    Parse a single property value / unit
    To determine the value that is to be used in nfo

    @param prop_info: A dictionary with property information
    @type prop_info: C{dict}

    @param value: Value of the property
    @type value: L{Expression}

    @param unit: Unit of the property value (e.g. km/h)
    @type unit: L{Unit} or C{None}

    @param size_bit: Bit that indicates the size of a multitile house
                     Set iff the item is a house
    @type size_bit: C{int} or C{None}

    @return: List of values to actually use (in nfo) for the property
    @rtype: L{Expression}
    """
    # Change value to use, except when the 'nfo' unit is used
    if unit is None or unit.type != "nfo":
        # Save the original value to test conversion against it
        org_value = value

        # Multiply by property-specific conversion factor
        mul, div = 1, 1
        if "unit_conversion" in prop_info:
            mul = prop_info["unit_conversion"]
            if isinstance(mul, tuple):
                mul, div = mul

        # Divide by conversion factor specified by unit
        if unit is not None:
            if "unit_type" not in prop_info or unit.type != prop_info["unit_type"]:
                raise generic.ScriptError("Invalid unit for property", value.pos)
            unit_mul, unit_div = unit.convert, 1
            if isinstance(unit_mul, tuple):
                unit_mul, unit_div = unit_mul
            mul *= unit_div
            div *= unit_mul

        # Factor out common factors
        gcd = generic.greatest_common_divisor(mul, div)
        mul //= gcd
        div //= gcd

        if isinstance(value, (expression.ConstantNumeric, expression.ConstantFloat)):
            # Even if mul == div == 1, we have to round floats and adjust value
            value = expression.ConstantNumeric(int(float(value.value) * mul / div + 0.5), value.pos)
            if unit is not None and "adjust_value" in prop_info:
                value = adjust_value(value, org_value, unit, prop_info["adjust_value"])
        elif mul != div:
            # Compute (value * mul + div/2) / div
            value = nmlop.MUL(value, mul)
            value = nmlop.ADD(value, int(div / 2))
            value = nmlop.DIV(value, div)

    elif isinstance(value, expression.ConstantFloat):
        # Round floats to ints
        value = expression.ConstantNumeric(int(value.value + 0.5), value.pos)

    # Apply value_function if it exists
    if "value_function" in prop_info:
        value = prop_info["value_function"](value)

    # Make multitile houses work
    if size_bit is not None:
        num_ids = house_sizes[size_bit]
        assert "multitile_function" in prop_info
        ret = prop_info["multitile_function"](value, num_ids, size_bit)
        assert len(ret) == num_ids
        return ret
    else:
        return [value]


def parse_property(prop_info, value_list, feature, id):
    """
    Parse a single property

    @param prop_info: A dictionary with property information
    @type prop_info: C{dict}

    @param value_list: List of values for the property, with unit conversion applied
    @type value_list: C{list} of L{Expression}

    @param feature: Feature of the associated item
    @type feature: C{int}

    @param id: ID of the associated item
    @type id: L{Expression}

    @return: A tuple containing the following:
                - List of properties to add to the action 0
                - List of actions to prepend
                - List of modifications to apply via action 6
                - List of actions to append
    @rtype: C{tuple} of (C{list} of L{Action0Property},
                         C{list} of L{BaseAction},
                         C{list} of 3-C{tuple},
                         C{list} of L{BaseAction})
    """
    action_list = []
    action_list_append = []
    mods = []

    if "custom_function" in prop_info:
        props = prop_info["custom_function"](*value_list)
    else:

        def apply_threshold(value):
            if not isinstance(value, tuple):
                return value
            assert len(value) == 3
            threshold, below, above = value
            return below if id.value < threshold else above

        # First process each element in the value_list
        final_values = []
        for i, value in enumerate(value_list):
            if "string_literal" in prop_info and (
                isinstance(value, expression.StringLiteral) or prop_info["string_literal"] != 4
            ):
                # Parse non-string expressions just like integers. User will have to take care of proper value.
                # This can be used to set a label (=string of length 4) to the value of a parameter.
                if not isinstance(value, expression.StringLiteral):
                    raise generic.ScriptError(
                        "Value for property {:d} must be a string literal".format(prop_info["num"]), value.pos
                    )
                if grfstrings.get_string_size(value.value, False, True) != prop_info["string_literal"]:
                    raise generic.ScriptError(
                        "Value for property {:d} must be of length {:d}".format(
                            prop_info["num"], prop_info["string_literal"]
                        ),
                        value.pos,
                    )

            elif isinstance(value, expression.ConstantNumeric):
                pass

            elif isinstance(value, expression.Parameter) and isinstance(value.num, expression.ConstantNumeric):
                mods.append((value.num.value, prop_info["size"], i * prop_info["size"] + 1))
                value = expression.ConstantNumeric(0)

            elif isinstance(value, expression.String):
                if "string" not in prop_info:
                    raise generic.ScriptError(
                        "String used as value for non-string property: " + str(prop_info["num"]), value.pos
                    )
                string_range = apply_threshold(prop_info["string"])
                stringid, string_actions = action4.get_string_action4s(feature, string_range, value, id)
                value = expression.ConstantNumeric(stringid)
                action_list_append.extend(string_actions)

            else:
                tmp_param, tmp_param_actions = actionD.get_tmp_parameter(value)
                mods.append((tmp_param, prop_info["size"], i * prop_info["size"] + 1))
                action_list.extend(tmp_param_actions)
                value = expression.ConstantNumeric(0)

            final_values.append(value)

        # Now, write a single Action0 Property with all of these values
        prop_num = apply_threshold(prop_info["num"])
        if prop_num != -1:
            props = [Action0Property(prop_num, final_values, prop_info["size"])]
        else:
            props = []

    return (props, action_list, mods, action_list_append)


def validate_prop_info_list(prop_info_list, pos_list, feature):
    """
    Perform various checks on a list of properties in a property-block
        - make sure that properties that should appear first (substitute type) appear first
        - make sure that required properties are set

    @param prop_info_list: List of dictionaries with property information
    @type prop_info_list: C{list} of C{dict}

    @param pos_list: List containing position information (order matches prop_info_list)
    @type pos_list: C{list} of L{Position}

    @param feature: Feature of the associated item
    @type feature: C{int}
    """
    first_warnings = [(info, pos_list[i]) for i, info in enumerate(prop_info_list) if "first" in info and i != 0]
    for prop_name, prop_info in properties[feature].items():
        if not isinstance(prop_info, list):
            prop_info = [prop_info]
        for info, pos in first_warnings:
            if info in prop_info:
                generic.print_warning(
                    generic.Warning.GENERIC,
                    "Property '{}' should be set before all other properties and graphics.".format(prop_name),
                    pos,
                )
                break
        for info in prop_info:
            if "required" in info and info not in prop_info_list:
                generic.print_error(
                    "Property '{}' is not set. Item will be invalid.".format(prop_name),
                    pos_list[-1],
                )


def parse_property_block(prop_list, feature, id, size):
    """
    Parse a property block to an action0 (with possibly various other actions)

    @param prop_list: List of properties to parse
    @type prop_list: C{list} of L{Property}

    @param feature: Feature of the associated item
    @type feature: C{int}

    @param id: ID of the associated item
    @type id: L{Expression}

    @param size: Size (for houses only)
    @type size: L{ConstantNumeric} or C{None}

    @return: List of resulting actions
    @rtype: C{list} of L{BaseAction}
    """
    action6.free_parameters.save()
    action_list = []
    action_list_append = []
    act6 = action6.Action6()

    action0, offset = create_action0(feature, id, act6, action_list)
    if feature == 0x07:
        size_bit = size.value if size is not None else 0
        action0.num_ids = house_sizes[size_bit]
    else:
        size_bit = None
        action0.num_ids = 1

    prop_info_list = []
    value_list_list = []
    pos_list = []
    for prop in prop_list:
        new_prop_info_list = get_property_info_list(feature, prop.name)
        prop_info_list.extend(new_prop_info_list)
        value_list_list.extend(
            parse_property_value(prop_info, prop.value, prop.unit, size_bit) for prop_info in new_prop_info_list
        )
        pos_list.extend(prop.name.pos for i in new_prop_info_list)

    validate_prop_info_list(prop_info_list, pos_list, feature)

    for prop_info, value_list in zip(prop_info_list, value_list_list):
        if "test_function" in prop_info and not prop_info["test_function"](*value_list):
            continue
        props, extra_actions, mods, extra_append_actions = parse_property(prop_info, value_list, feature, id)
        action_list.extend(extra_actions)
        action_list_append.extend(extra_append_actions)
        for mod in mods:
            act6.modify_bytes(mod[0], mod[1], mod[2] + offset)
        for p in props:
            offset += p.get_size()
        action0.prop_list.extend(props)

    if len(act6.modifications) > 0:
        action_list.append(act6)
    if len(action0.prop_list) != 0:
        action_list.append(action0)

    action_list.extend(action_list_append)

    action6.free_parameters.restore()
    return action_list


class IDListProp(BaseAction0Property):
    def __init__(self, prop_num, id_list):
        self.prop_num = prop_num
        self.id_list = id_list

    def write(self, file):
        file.print_bytex(self.prop_num)
        for i, id_val in enumerate(self.id_list):
            if i > 0 and i % 5 == 0:
                file.newline()
            file.print_string(id_val.value, False, True)
        file.newline()

    def get_size(self):
        return len(self.id_list) * 4 + 1


def get_cargolist_action(cargo_list):
    action0 = Action0(0x08, 0)
    action0.prop_list.append(IDListProp(0x09, cargo_list))
    action0.num_ids = len(cargo_list)
    return [action0]


def get_tracktypelist_action(table_prop_id, cond_tracktype_not_defined, tracktype_list):
    action6.free_parameters.save()
    act6 = action6.Action6()

    action_list = []
    action0, offset = create_action0(0x08, expression.ConstantNumeric(0), act6, action_list)
    id_table = []
    offset += 1  # Skip property number
    for tracktype in tracktype_list:
        if isinstance(tracktype, expression.StringLiteral):
            id_table.append(tracktype)
            offset += 4
            continue
        param, extra_actions = actionD.get_tmp_parameter(
            expression.ConstantNumeric(expression.parse_string_to_dword(tracktype[-1]))
        )
        action_list.extend(extra_actions)
        for idx in range(len(tracktype) - 2, -1, -1):
            val = expression.ConstantNumeric(expression.parse_string_to_dword(tracktype[idx]))
            action_list.append(action7.SkipAction(0x09, 0x00, 4, (cond_tracktype_not_defined, None), val.value, 1))
            action_list.append(
                actionD.ActionD(
                    expression.ConstantNumeric(param),
                    expression.ConstantNumeric(0xFF),
                    nmlop.ASSIGN,
                    expression.ConstantNumeric(0xFF),
                    val,
                )
            )
        act6.modify_bytes(param, 4, offset)
        id_table.append(expression.StringLiteral(r"\00\00\00\00", None))
        offset += 4
    action0.prop_list.append(IDListProp(table_prop_id, id_table))
    action0.num_ids = len(tracktype_list)

    if len(act6.modifications) > 0:
        action_list.append(act6)

    action_list.append(action0)
    action6.free_parameters.restore()
    return action_list


class ByteListProp(BaseAction0Property):
    def __init__(self, prop_num, data):
        self.prop_num = prop_num
        self.data = data

    def write(self, file):
        file.print_bytex(self.prop_num)
        file.newline()
        for i, data_val in enumerate(self.data):
            if i > 0 and i % 8 == 0:
                file.newline()
            file.print_bytex(ord(data_val))
        file.newline()

    def get_size(self):
        return len(self.data) + 1


def get_snowlinetable_action(snowline_table):
    assert len(snowline_table) == 12 * 32

    action6.free_parameters.save()
    action_list = []
    tmp_param_map = {}  # Cache for tmp parameters
    act6 = action6.Action6()

    act0, offset = create_action0(0x08, expression.ConstantNumeric(0), act6, action_list)
    act0.num_ids = 1
    offset += 1  # Skip property number

    data_table = []
    idx = 0
    while idx < len(snowline_table):
        val = snowline_table[idx]
        if isinstance(val, expression.ConstantNumeric):
            data_table.append(val.value)
            idx += 1
            continue

        if idx + 3 >= len(snowline_table):
            tmp_param, tmp_param_actions = actionD.get_tmp_parameter(val)
            tmp_param_map[val] = tmp_param
            act6.modify_bytes(tmp_param, 1, offset + idx)
            action_list.extend(tmp_param_actions)
            data_table.append(0)
            idx += 1
            continue

        # Merge the next 4 values together in a single parameter.
        val2 = nmlop.SHIFT_LEFT(snowline_table[idx + 1], 8)
        val3 = nmlop.SHIFT_LEFT(snowline_table[idx + 2], 16)
        val4 = nmlop.SHIFT_LEFT(snowline_table[idx + 3], 24)
        expr = nmlop.OR(val, val2)
        expr = nmlop.OR(expr, val3)
        expr = nmlop.OR(expr, val4)
        expr = expr.reduce()

        # Cache lookup, saves some ActionDs
        if expr in tmp_param_map:
            tmp_param, tmp_param_actions = tmp_param_map[expr], []
        else:
            tmp_param, tmp_param_actions = actionD.get_tmp_parameter(expr)
            tmp_param_map[expr] = tmp_param

        act6.modify_bytes(tmp_param, 4, offset + idx)
        action_list.extend(tmp_param_actions)
        data_table.extend([0, 0, 0, 0])
        idx += 4

    act0.prop_list.append(ByteListProp(0x10, "".join([chr(x) for x in data_table])))
    if len(act6.modifications) > 0:
        action_list.append(act6)
    action_list.append(act0)
    action6.free_parameters.restore()
    return action_list


def get_basecost_action(basecost):
    action6.free_parameters.save()
    action_list = []
    tmp_param_map = {}  # Cache for tmp parameters

    # We want to avoid writing lots of action0s if possible
    i = 0
    while i < len(basecost.costs):
        cost = basecost.costs[i]
        act6 = action6.Action6()

        act0, offset = create_action0(0x08, cost.name, act6, action_list)
        first_id = cost.name.value if isinstance(cost.name, expression.ConstantNumeric) else None

        num_ids = 1  # Number of values that will be written in one go
        values = []
        # try to capture as many values as possible
        while True:
            cost = basecost.costs[i]
            if isinstance(cost.value, expression.ConstantNumeric):
                values.append(cost.value)
            else:
                # Cache lookup, saves some ActionDs
                if cost.value in tmp_param_map:
                    tmp_param, tmp_param_actions = tmp_param_map[cost.value], []
                else:
                    tmp_param, tmp_param_actions = actionD.get_tmp_parameter(cost.value)
                    tmp_param_map[cost.value] = tmp_param
                act6.modify_bytes(tmp_param, 1, offset + num_ids)
                action_list.extend(tmp_param_actions)
                values.append(expression.ConstantNumeric(0))

            # check if we can append the next to this one (it has to be consecutively numbered)
            if first_id is not None and (i + 1) < len(basecost.costs):
                nextcost = basecost.costs[i + 1]
                if isinstance(nextcost.name, expression.ConstantNumeric) and nextcost.name.value == first_id + num_ids:
                    num_ids += 1
                    i += 1
                    # Yes We Can, continue the loop to append this value to the list and try further
                    continue
            # No match, so stop it and write an action0
            break

        act0.prop_list.append(Action0Property(0x08, values, 1))
        act0.num_ids = num_ids
        if len(act6.modifications) > 0:
            action_list.append(act6)
        action_list.append(act0)
        i += 1
    action6.free_parameters.restore()
    return action_list


class LanguageTranslationTable(BaseAction0Property):
    def __init__(self, num, name_list, extra_names):
        self.num = num
        self.mappings = []
        for name, idx in name_list.items():
            self.mappings.append((idx, name))
            if name in extra_names:
                for extra_name in extra_names[name]:
                    self.mappings.append((idx, extra_name))
        # Sort to keep the output deterministic
        self.mappings.sort()

    def write(self, file):
        file.print_bytex(self.num)
        for mapping in self.mappings:
            file.print_bytex(mapping[0])
            file.print_string(mapping[1])
        file.print_bytex(0)
        file.newline()

    def get_size(self):
        size = 2
        for mapping in self.mappings:
            size += 1 + grfstrings.get_string_size(mapping[1])
        return size


def get_language_translation_tables(lang):
    action0 = Action0(0x08, lang.langid)
    if lang.genders is not None:
        action0.prop_list.append(LanguageTranslationTable(0x13, lang.genders, lang.gender_map))
    if lang.cases is not None:
        action0.prop_list.append(LanguageTranslationTable(0x14, lang.cases, lang.case_map))
    if lang.plural is not None:
        action0.prop_list.append(Action0Property(0x15, expression.ConstantNumeric(lang.plural), 1))
    if len(action0.prop_list) > 0:
        return [action0]
    return []


disable_info = {
    # Vehicles: set climates_available to 0
    0x00: {"num": 116, "props": [{"num": 0x06, "size": 1, "value": 0}]},
    0x01: {"num": 88, "props": [{"num": 0x06, "size": 1, "value": 0}]},
    0x02: {"num": 11, "props": [{"num": 0x06, "size": 1, "value": 0}]},
    0x03: {"num": 41, "props": [{"num": 0x06, "size": 1, "value": 0}]},
    # Houses / industries / airports: Set substitute_type to FF
    0x07: {"num": 110, "props": [{"num": 0x08, "size": 1, "value": 0xFF}]},
    0x0A: {"num": 37, "props": [{"num": 0x08, "size": 1, "value": 0xFF}]},
    0x0D: {"num": 10, "props": [{"num": 0x08, "size": 1, "value": 0xFF}]},
    # Cargos: Set bitnum to FF and label to 0
    0x0B: {"num": 27, "props": [{"num": 0x08, "size": 1, "value": 0xFF}, {"num": 0x17, "size": 4, "value": 0}]},
}


def get_disable_actions(disable):
    """
    Get the action list for a disable_item block

    @param disable: Disable block
    @type disable: L{DisableItem}

    @return: A list of resulting actions
    @rtype: C{list} of L{BaseAction}
    """
    feature = disable.feature.value
    if feature not in disable_info:
        raise generic.ScriptError("disable_item() is not available for feature {:d}.".format(feature), disable.pos)
    if disable.first_id is None:
        # No ids set -> disable all
        assert disable.last_id is None
        first = 0
        num = disable_info[feature]["num"]
    else:
        first = disable.first_id.value
        if disable.last_id is None:
            num = 1
        else:
            num = disable.last_id.value - first + 1

    act0 = Action0(feature, first)
    act0.num_ids = num
    for prop in disable_info[feature]["props"]:
        act0.prop_list.append(
            Action0Property(prop["num"], num * [expression.ConstantNumeric(prop["value"])], prop["size"])
        )

    return [act0]


class EngineOverrideProp(BaseAction0Property):
    def __init__(self, source, target):
        self.source = source
        self.target = target

    def write(self, file):
        file.print_bytex(0x11)
        file.print_dwordx(self.source)
        file.print_dwordx(self.target)
        file.newline()

    def get_size(self):
        return 9


def get_engine_override_action(override):
    act0 = Action0(0x08, 0)
    act0.num_ids = 1
    act0.prop_list.append(EngineOverrideProp(override.source_grfid, override.grfid))
    return [act0]


def parse_sort_block(feature, vehid_list):
    prop_num = [0x1A, 0x20, 0x1B, 0x1B]
    action_list = []
    if len(vehid_list) >= 2:
        last = vehid_list[0]
        idx = len(vehid_list) - 1
        while idx >= 0:
            cur = vehid_list[idx]
            prop = Action0Property(prop_num[feature], [last], 3)
            action_list.append(Action0(feature, cur.value))
            action_list[-1].prop_list.append(prop)
            last = cur
            idx -= 1
    return action_list


callback_flag_properties = {
    0x00: two_byte_property(0x1E, 0x31),
    0x01: two_byte_property(0x17, 0x28),
    0x02: two_byte_property(0x12, 0x22),
    0x03: two_byte_property(0x14, 0x22),
    0x04: {"size": 1, "num": 0x0B},
    0x05: {"size": 1, "num": 0x08},
    0x07: two_byte_property(0x14, 0x1D),
    0x09: {"size": 1, "num": 0x0E},
    0x0A: two_byte_property(0x21, 0x22),
    0x0B: {"size": 1, "num": 0x1A},
    0x0F: {"size": 2, "num": 0x15},
    0x11: {"size": 1, "num": 0x0E},
    0x14: {"size": 1, "num": 0x11},
}


def get_callback_flags_actions(feature, id, flags):
    """
    Get a list of actions to set the callback flags of a certain item

    @param feature: Feature of the item
    @type feature: C{int}

    @param id: ID of the item
    @type id: L{Expression}

    @param flags: Value of the 'callback_flags' property
    @type flags: C{int}

    @return: A list of actions
    @rtype: C{list} of L{BaseAction}
    """
    assert isinstance(id, expression.ConstantNumeric)
    act0, offset = create_action0(feature, id, None, None)
    act0.num_ids = 1
    assert feature in callback_flag_properties

    prop_info_list = callback_flag_properties[feature]
    if not isinstance(prop_info_list, list):
        prop_info_list = [prop_info_list]
    for prop_info in prop_info_list:
        parsed_value = parse_property_value(prop_info, expression.ConstantNumeric(flags))
        if parsed_value[0].value != 0:
            act0.prop_list.append(Action0Property(prop_info["num"], parsed_value, prop_info["size"]))

    return [act0]


def get_volume_actions(volume_list):
    """
    Get a list of actions to set sound volumes

    @param volume_list: List of (sound id, volume) tuples, sorted in ascending order
    @type volume_list: C{list} of (C{int}, C{int})-tuples

    @return: A list of actions
    @rtype: C{list} of L{BaseAction}
    """
    action_list = []
    first_id = None  # First ID in a series of consecutive IDs
    value_series = []  # Series of values to write in a single action

    for id, volume in volume_list:
        if first_id is None:
            first_id = id
        continue_series = first_id + len(value_series) == id
        if continue_series:
            value_series.append(volume)

        if not continue_series or id == volume_list[-1][0]:
            # Write action for this series
            act0, offset = create_action0(0x0C, expression.ConstantNumeric(first_id), None, None)
            act0.num_ids = len(value_series)
            act0.prop_list.append(Action0Property(0x08, [expression.ConstantNumeric(v) for v in value_series], 1))
            action_list.append(act0)

            # start new series, if needed
            if not continue_series:
                first_id = id
                value_series = [volume]

    return action_list


class StationLayoutProp(BaseAction0Property):
    def __init__(self, layouts):
        self.layouts = layouts

    def write(self, file):
        file.print_bytex(0x1A)
        file.print_byte(len(self.layouts))
        file.newline()
        for layout in self.layouts:
            layout.write(file)

    def get_size(self):
        size = 2
        for layout in self.layouts:
            size += layout.get_size()
        return size


def get_layout_action0(feature, id, layouts):
    action_list = []
    act0, offset = create_action0(feature, id, None, None)
    action6.free_parameters.save()
    act6 = action6.Action6()

    for layout in layouts:
        layout.write_action_value(action_list, act6, offset)

    if len(act6.modifications) > 0:
        action_list.append(act6)

    act0.num_ids = 1
    act0.prop_list.append(StationLayoutProp(layouts))
    action_list.append(act0)

    action6.free_parameters.restore()
    return action_list


def get_copy_layout_action0(feature, id, source_id):
    act0, offset = create_action0(feature, id, None, None)
    act0.num_ids = 1
    act0.prop_list.append(Action0Property(0x0A, source_id, 1 if source_id.value < 0xFF else 3))
    return [act0]



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic
from nml.actions import actionB
from nml.ast import base_statement


class Error(base_statement.BaseStatement):
    """
    An error has occurred while parsing the GRF. This can be anything ranging from
    an incompatible GRF file that was found or a game setting that is set to the
    wrong value to a wrong combination of parameters. The action taken by the host
    depends on the severity level of the error.
    NML equivalent: error(level, message[, extra_text[, parameter1[, parameter2]]]).

    @ivar params: Extra expressions whose value can be used in the error string.
    @type params: C{list} of L{Expression}

    @ivar severity: Severity level of this error, value between 0 and 3.
    @type severity: L{Expression}

    @ivar msg: The string to be used for this error message. This can be either
               one of the predefined error strings or a custom string from the
               language file.
    @type msg: L{Expression}

    @ivar data: Optional extra message that is inserted in place of the second
                {STRING}-code of msg.
    @type data: C{None} or L{String} or L{StringLiteral}
    """

    def __init__(self, param_list, pos):
        base_statement.BaseStatement.__init__(self, "error()", pos)
        if not 2 <= len(param_list) <= 5:
            raise generic.ScriptError(
                "'error' expects between 2 and 5 parameters, got " + str(len(param_list)), self.pos
            )
        self.severity = param_list[0]
        self.msg = param_list[1]
        self.data = param_list[2] if len(param_list) >= 3 else None
        self.params = param_list[3:]

    def pre_process(self):
        self.severity = self.severity.reduce([actionB.error_severity])
        self.msg = self.msg.reduce([actionB.default_error_msg])
        if self.data:
            self.data = self.data.reduce()
        self.params = [x.reduce() for x in self.params]

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Error message")
        generic.print_dbg(indentation + 2, "Message:")
        self.msg.debug_print(indentation + 4)

        generic.print_dbg(indentation + 2, "Severity:")
        self.severity.debug_print(indentation + 4)

        generic.print_dbg(indentation, "Data: ")
        if self.data is not None:
            self.data.debug_print(indentation + 4)

        if len(self.params) > 0:
            generic.print_dbg(indentation + 2, "Param1: ")
            self.params[0].debug_print(indentation + 4)
        if len(self.params) > 1:
            generic.print_dbg(indentation + 2, "Param2: ")
            self.params[1].debug_print(indentation + 4)

    def get_action_list(self):
        return actionB.parse_error_block(self)

    def __str__(self):
        sev = str(self.severity)
        if isinstance(self.severity, expression.ConstantNumeric):
            for s in actionB.error_severity:
                if self.severity.value == actionB.error_severity[s]:
                    sev = s
                    break
        res = "error({}, {}".format(sev, self.msg)
        if self.data is not None:
            res += ", {}".format(self.data)
        if len(self.params) > 0:
            res += ", {}".format(self.params[0])
        if len(self.params) > 1:
            res += ", {}".format(self.params[1])
        res += ");\n"
        return res



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic
from nml.actions import actionE
from nml.ast import base_statement


class DeactivateBlock(base_statement.BaseStatement):
    def __init__(self, grfid_list, pos):
        base_statement.BaseStatement.__init__(self, "deactivate()", pos)
        self.grfid_list = grfid_list

    def pre_process(self):
        # Parse (string-)expressions to integers
        self.grfid_list = [expression.parse_string_to_dword(grfid.reduce()) for grfid in self.grfid_list]

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Deactivate other newgrfs:")
        for grfid in self.grfid_list:
            grfid.debug_print(indentation + 2)

    def get_action_list(self):
        return actionE.parse_deactivate_block(self)

    def __str__(self):
        return "deactivate({});\n".format(", ".join(str(grfid) for grfid in self.grfid_list))



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

import array


def _encode(data):
    """
    GRF compression algorithm.

    @param data: Uncompressed data.
    @type  data: C{str}, C{bytearray}, C{bytes} or similar.

    @return: Compressed data.
    @rtype:  C{bytearray}
    """
    stream = data.tobytes()
    position = 0
    output = array.array("B")
    literal_bytes = array.array("B")
    stream_len = len(stream)

    while position < stream_len:
        overlap_len = 0
        start_pos = max(0, position - (1 << 11) + 1)
        # Loop through the lookahead buffer.
        for i in range(3, min(stream_len - position + 1, 16)):
            # Set pattern to find the longest match.
            pattern = stream[position : position + i]
            # Find the pattern match in the window.
            result = stream.find(pattern, start_pos, position)
            # If match failed, we've found the longest.
            if result < 0:
                break

            p = position - result
            overlap_len = i
            start_pos = result

        if overlap_len > 0:
            if len(literal_bytes) > 0:
                output.append(len(literal_bytes))
                output.extend(literal_bytes)
                literal_bytes = array.array("B")

            val = ((-overlap_len) << 3) & 0xFF | (p >> 8)
            output.append(val)
            output.append(p & 0xFF)
            position += overlap_len
        else:
            literal_bytes.append(stream[position])
            if len(literal_bytes) == 0x80:
                output.append(0)
                output.extend(literal_bytes)
                literal_bytes = array.array("B")

            position += 1

    if len(literal_bytes) > 0:
        output.append(len(literal_bytes))
        output.extend(literal_bytes)

    return output


"""
True if the encoding is provided by a native module.
Used for verbose information.
"""
is_native = False

try:
    from nml_lz77 import encode  # lgtm[py/unused-import]

    is_native = True
except ImportError:
    encode = _encode



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, global_constants
from nml.ast import base_statement


class Constant(base_statement.BaseStatement):
    def __init__(self, name, value):
        base_statement.BaseStatement.__init__(self, "constant", name.pos, False, False)
        self.name = name
        self.value = value

    def register_names(self):
        if not isinstance(self.name, expression.Identifier):
            raise generic.ScriptError("Constant name should be an identifier", self.name.pos)
        if self.name.value in global_constants.constant_numbers:
            raise generic.ScriptError("Redefinition of constant '{}'.".format(self.name.value), self.name.pos)
        global_constants.constant_numbers[self.name.value] = self.value.reduce_constant(
            global_constants.const_list
        ).value

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Constant")
        generic.print_dbg(indentation + 2, "Name:")
        self.name.debug_print(indentation + 4)

        generic.print_dbg(indentation + 2, "Value:")
        self.value.debug_print(indentation + 4)

    def get_action_list(self):
        return []

    def __str__(self):
        return "const {} = {};".format(self.name, self.value)



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, global_constants
from nml.actions import action0
from nml.ast import base_statement


class EngineOverride(base_statement.BaseStatement):
    """
    AST Node for an engine override.

    @ivar grfid: GRFid of the grf to override the engines from.
    @type grfid: C{int}

    @ivar source_grfid: GRFid of the grf that overrides the engines.
    @type source_grfid: C{int}
    """

    def __init__(self, args, pos):
        base_statement.BaseStatement.__init__(self, "engine_override()", pos)
        self.args = args

    def pre_process(self):
        if len(self.args) not in (1, 2):
            raise generic.ScriptError("engine_override expects 1 or 2 parameters", self.pos)

        if len(self.args) == 1:
            try:
                self.source_grfid = expression.Identifier("GRFID").reduce(global_constants.const_list).value
                assert isinstance(self.source_grfid, int)
            except generic.ScriptError:
                raise generic.ScriptError("GRFID of this grf is required, but no grf-block is defined.", self.pos)
        else:
            self.source_grfid = expression.parse_string_to_dword(self.args[0].reduce(global_constants.const_list))

        self.grfid = expression.parse_string_to_dword(self.args[-1].reduce(global_constants.const_list))

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Engine override")
        generic.print_dbg(indentation, "Source:", self.source_grfid)
        generic.print_dbg(indentation, "Target:", self.grfid)

    def get_action_list(self):
        return action0.get_engine_override_action(self)

    def __str__(self):
        return "engine_override({});\n".format(", ".join(str(x) for x in self.args))



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, global_constants
from nml.actions import action2, action2random, action2var
from nml.ast import base_statement, general

var_ranges = {"SELF": 0x89, "PARENT": 0x8A}

# Used by Switch and RandomSwitch
switch_base_class = action2.make_sprite_group_class(False, True, True)


class Switch(switch_base_class):
    def __init__(self, param_list, body, pos):
        base_statement.BaseStatement.__init__(self, "switch-block", pos, False, False)
        if len(param_list) < 4:
            raise generic.ScriptError(
                "Switch-block requires at least 4 parameters, encountered " + str(len(param_list)), pos
            )
        if not isinstance(param_list[1], expression.Identifier):
            raise generic.ScriptError(
                "Switch-block parameter 2 'variable range' must be an identifier.", param_list[1].pos
            )
        if param_list[1].value in var_ranges:
            self.var_range = var_ranges[param_list[1].value]
        else:
            raise generic.ScriptError(
                "Unrecognized value for switch parameter 2 'variable range': '{}'".format(param_list[1].value),
                param_list[1].pos,
            )
        if not isinstance(param_list[2], expression.Identifier):
            raise generic.ScriptError("Switch-block parameter 3 'name' must be an identifier.", param_list[2].pos)
        self.initialize(param_list[2], general.parse_feature(param_list[0]).value, len(param_list) - 4)
        self.expr = param_list[-1]
        self.body = body
        self.param_list = param_list[3:-1]
        # register_map is a dict to be duck-compatible with spritelayouts.
        # But because feature_set has only one item for switches, register_map also has at most one item.
        self.register_map = {}

    def pre_process(self):
        # Check parameter names
        seen_names = set()
        for param in self.param_list:
            if not isinstance(param, expression.Identifier):
                raise generic.ScriptError("switch parameter names must be identifiers.", param.pos)
            if param.value in seen_names:
                raise generic.ScriptError("Duplicate parameter name '{}' encountered.".format(param.value), param.pos)
            seen_names.add(param.value)

        feature = next(iter(self.feature_set))
        var_scope = action2var.get_scope(feature, self.var_range)
        if var_scope is None:
            raise generic.ScriptError("Requested scope not available for this feature.", self.pos)

        # Allocate registers
        param_map = {}
        param_registers = []
        for param in self.param_list:
            reg = action2var.VarAction2CallParam(param.value)
            param_registers.append(reg)
            param_map[param.value] = reg
        param_map = (param_map, lambda name, value, pos: action2var.VarAction2LoadCallParam(value, name))
        self.register_map[feature] = param_registers

        self.expr = action2var.reduce_varaction2_expr(self.expr, var_scope, [param_map])
        self.body.reduce_expressions(var_scope, [param_map])
        switch_base_class.pre_process(self)

    def optimise(self):
        if self.optimised:
            return self.optimised is not self

        # Constant condition, pick the right result
        if isinstance(self.expr, expression.ConstantNumeric):
            for r in self.body.ranges[:]:
                if r.min.value <= self.expr.value <= r.max.value:
                    self.optimised = r.result.value if r.result.value else self.expr
            if not self.optimised and self.body.default:
                self.optimised = self.body.default.value if self.body.default.value else self.expr

        # Default result only
        if not self.optimised and self.expr.is_read_only() and self.body.default and len(self.body.ranges) == 0:
            self.optimised = self.body.default.value

        # If we return an expression, just rewrite ourself to keep the correct scope
        # so a return action with the wrong scope doesn't need to be created later
        if (
            self.optimised
            and not isinstance(self.optimised, expression.ConstantNumeric)
            and not (isinstance(self.optimised, expression.SpriteGroupRef) and not self.optimised.is_procedure)
            and not isinstance(self.optimised, expression.String)
        ):
            self.expr = self.optimised
            self.body.ranges = []
            self.body.default = SwitchValue(None, True, self.optimised.pos)
            self.optimised = self

        if self.optimised:
            generic.print_warning(
                generic.Warning.OPTIMISATION,
                "Block '{}' returns a constant, optimising.".format(self.name.value),
                self.pos,
            )
            return self.optimised is not self

        self.optimised = self  # Prevent multiple run on the same non-optimisable Switch
        return False

    def collect_references(self):
        all_refs = self.expr.collect_references()
        for result in [r.result for r in self.body.ranges] + [self.body.default]:
            if result is not None and result.value is not None:
                all_refs += result.value.collect_references()
        return all_refs

    def is_read_only(self):
        for result in [r.result for r in self.body.ranges] + [self.body.default]:
            if result is not None and result.value is not None:
                if not result.value.is_read_only():
                    return False
        return self.expr.is_read_only()

    def debug_print(self, indentation):
        generic.print_dbg(
            indentation, "Switch, Feature = {:d}, name = {}".format(next(iter(self.feature_set)), self.name.value)
        )

        if self.param_list:
            generic.print_dbg(indentation + 2, "Parameters:")
            for param in self.param_list:
                param.debug_print(indentation + 4)

        generic.print_dbg(indentation + 2, "Expression:")
        self.expr.debug_print(indentation + 4)

        generic.print_dbg(indentation + 2, "Body:")
        self.body.debug_print(indentation + 4)

    def get_action_list(self):
        if self.prepare_act2_output():
            return action2var.parse_varaction2(self)
        return []

    def __str__(self):
        var_range = "SELF" if self.var_range == 0x89 else "PARENT"
        params = "" if not self.param_list else "{}, ".format(", ".join(str(x) for x in self.param_list))
        return "switch({}, {}, {}, {}{}) {{\n{}}}\n".format(
            next(iter(self.feature_set)), var_range, self.name, params, self.expr, self.body
        )


class SwitchBody:
    """
    AST-node representing the body of a switch block
    This contains the various ranges as well as the default value

    @ivar ranges: List of ranges
    @type ranges: C{list} of L{SwitchRange}

    @ivar default: Default result to use if no range matches
    @type default: L{SwitchValue} or C{None} if N/A
    """

    def __init__(self, ranges, default):
        self.ranges = ranges
        self.default = default

    def reduce_expressions(self, var_scope, extra_dicts=None):
        if extra_dicts is None:
            extra_dicts = []
        for r in self.ranges[:]:
            if r.min is r.max and isinstance(r.min, expression.Identifier) and r.min.value == "default":
                if self.default is not None:
                    raise generic.ScriptError(
                        "Switch-block has more than one default, which is impossible.", r.result.pos
                    )
                self.default = r.result
                self.ranges.remove(r)
            else:
                r.reduce_expressions(var_scope, extra_dicts)
        if self.default is not None:
            if self.default.value is not None:
                self.default.value = action2var.reduce_varaction2_expr(self.default.value, var_scope, extra_dicts)
            if len(self.ranges) != 0:
                if any(self.default.value != r.result.value for r in self.ranges):
                    return
                generic.print_warning(
                    generic.Warning.OPTIMISATION,
                    "Switch-Block ranges are the same as default, optimising.",
                    self.default.pos,
                )
                self.ranges = []

    def debug_print(self, indentation):
        for r in self.ranges:
            r.debug_print(indentation)
        if self.default is not None:
            generic.print_dbg(indentation, "Default:")
            self.default.debug_print(indentation + 2)

    def __str__(self):
        ret = "".join("\t{}\n".format(r) for r in self.ranges)
        if self.default is not None:
            ret += "\t{}\n".format(str(self.default))
        return ret


class SwitchRange:
    def __init__(self, min, max, result, unit=None):
        self.min = min
        self.max = max
        self.result = result
        self.unit = unit

    def reduce_expressions(self, var_scope, extra_dicts=None):
        if extra_dicts is None:
            extra_dicts = []
        self.min = self.min.reduce(global_constants.const_list)
        self.max = self.max.reduce(global_constants.const_list)
        if self.result.value is not None:
            self.result.value = action2var.reduce_varaction2_expr(self.result.value, var_scope, extra_dicts)

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Min:")
        self.min.debug_print(indentation + 2)
        generic.print_dbg(indentation, "Max:")
        self.max.debug_print(indentation + 2)
        generic.print_dbg(indentation, "Result:")
        self.result.debug_print(indentation + 2)

    def __str__(self):
        ret = str(self.min)
        if self.min is not self.max and (
            not isinstance(self.min, expression.ConstantNumeric)
            or not isinstance(self.max, expression.ConstantNumeric)
            or self.max.value != self.min.value
        ):
            ret += ".." + str(self.max)
        ret += ": " + str(self.result)
        return ret


class SwitchValue:
    """
    Class representing a single returned value or sprite group in a switch-block
    Also used for random-switch and graphics blocks

    @ivar value: Value to return
    @type value: L{Expression} or C{None}

    @ivar is_return: Whether the return keyword was present
    @type is_return: C{bool}

    @ivar pos: Position information
    @type pos: L{Position}
    """

    def __init__(self, value, is_return, pos):
        self.value = value
        self.is_return = is_return
        self.pos = pos

    def debug_print(self, indentation):
        if self.value is None:
            assert self.is_return
            generic.print_dbg(indentation, "Return computed value")
        else:
            generic.print_dbg(indentation, "Return value:" if self.is_return else "Go to block:")
            self.value.debug_print(indentation + 2)

    def __str__(self):
        if self.value is None:
            assert self.is_return
            return "return;"
        elif self.is_return:
            return "return {};".format(self.value)
        else:
            return "{};".format(self.value)


class RandomSwitch(switch_base_class):
    def __init__(self, param_list, choices, pos):
        base_statement.BaseStatement.__init__(self, "random_switch-block", pos, False, False)
        if not (3 <= len(param_list) <= 4):
            raise generic.ScriptError(
                "random_switch requires 3 or 4 parameters, encountered {:d}".format(len(param_list)), pos
            )
        # feature
        feature = general.parse_feature(param_list[0]).value

        # type
        self.type = param_list[1]
        # Extract type name and possible argument
        if isinstance(self.type, expression.Identifier):
            self.type_count = None
        elif isinstance(self.type, expression.FunctionCall):
            if len(self.type.params) == 0:
                self.type_count = None
            elif len(self.type.params) == 1:
                # var_scope is really weird for type=BACKWARD/FORWARD.
                var_scope = action2var.get_scope(feature, 0x89)
                self.type_count = action2var.reduce_varaction2_expr(self.type.params[0], var_scope)
            else:
                raise generic.ScriptError(
                    "Value for random_switch parameter 2 'type' can have only one parameter.", self.type.pos
                )
            self.type = self.type.name
        else:
            raise generic.ScriptError(
                "random_switch parameter 2 'type' should be an identifier, possibly with a parameter.", self.type.pos
            )

        # name
        if not isinstance(param_list[2], expression.Identifier):
            raise generic.ScriptError("random_switch parameter 3 'name' should be an identifier", pos)
        name = param_list[2]

        # triggers
        self.triggers = param_list[3] if len(param_list) == 4 else expression.ConstantNumeric(0)

        # body
        self.choices = []
        self.dependent = []
        self.independent = []
        for choice in choices:
            if isinstance(choice.probability, expression.Identifier):
                if choice.probability.value == "dependent":
                    self.dependent.append(choice.result)
                    continue
                elif choice.probability.value == "independent":
                    self.independent.append(choice.result)
                    continue
            self.choices.append(choice)
        if len(self.choices) == 0:
            raise generic.ScriptError("random_switch requires at least one possible choice", pos)

        self.initialize(name, feature)
        self.random_act2 = None  # Set during action generation to resolve dependent/independent chains

    def pre_process(self):
        feature = next(iter(self.feature_set))
        # var_scope is really weird for type=BACKWARD/FORWARD.
        # Expressions in cases will still refer to the origin vehicle.
        var_scope = action2var.get_scope(feature, 0x8A if self.type.value == "PARENT" else 0x89)

        for choice in self.choices:
            choice.reduce_expressions(var_scope)

        for dep_list in (self.dependent, self.independent):
            for i, dep in enumerate(dep_list[:]):
                if dep.is_return:
                    raise generic.ScriptError(
                        "Expected a random_switch identifier after (in)dependent, not a return.", dep.pos
                    )
                dep_list[i] = dep.value.reduce(global_constants.const_list)
                # Make sure, all [in]dependencies refer to existing random switch blocks
                if (not isinstance(dep_list[i], expression.SpriteGroupRef)) or len(dep_list[i].param_list) > 0:
                    raise generic.ScriptError("Value for (in)dependent should be an identifier", dep_list[i].pos)
                spritegroup = action2.resolve_spritegroup(dep_list[i].name)
                if not isinstance(spritegroup, RandomSwitch):
                    raise generic.ScriptError(
                        "Value of (in)dependent '{}' should refer to a random_switch.".format(dep_list[i].name.value),
                        dep_list[i].pos,
                    )

        self.triggers = self.triggers.reduce_constant(global_constants.const_list)
        if not (0 <= self.triggers.value <= 255):
            raise generic.ScriptError(
                "random_switch parameter 4 'triggers' out of range 0..255, encountered " + str(self.triggers.value),
                self.triggers.pos,
            )

        switch_base_class.pre_process(self)

    def optimise(self):
        if self.optimised:
            return self.optimised is not self

        # Triggers have side effects, and can't be skipped.
        # Scope for expressions can be different in referencing location, so don't optimise them.
        if self.triggers.value == 0 and len(self.choices) == 1:
            optimised = self.choices[0].result.value
            if (
                isinstance(optimised, expression.ConstantNumeric)
                or (isinstance(optimised, expression.SpriteGroupRef) and not optimised.is_procedure)
                or isinstance(optimised, expression.String)
            ):
                generic.print_warning(
                    generic.Warning.OPTIMISATION,
                    "Block '{}' returns a constant, optimising.".format(self.name.value),
                    self.pos,
                )
                self.optimised = optimised
                return True

        self.optimised = self  # Prevent multiple run on the same non-optimisable RandomSwitch
        return False

    def collect_references(self):
        all_refs = []
        for choice in self.choices:
            all_refs += choice.result.value.collect_references()
        return all_refs

    def is_read_only(self):
        for choice in self.choices:
            if not choice.result.value.is_read_only():
                return False
        return True

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Random")
        generic.print_dbg(indentation + 2, "Feature:", next(iter(self.feature_set)))
        generic.print_dbg(indentation + 2, "Type:")
        self.type.debug_print(indentation + 4)
        generic.print_dbg(indentation + 2, "Name:", self.name.value)

        generic.print_dbg(indentation + 2, "Triggers:")
        self.triggers.debug_print(indentation + 4)
        for dep in self.dependent:
            generic.print_dbg(indentation + 2, "Dependent on:")
            dep.debug_print(indentation + 4)
        for indep in self.independent:
            generic.print_dbg(indentation + 2, "Independent from:")
            indep.debug_print(indentation + 4)

        generic.print_dbg(indentation + 2, "Choices:")
        for choice in self.choices:
            choice.debug_print(indentation + 4)

    def get_action_list(self):
        if self.prepare_act2_output():
            return action2random.parse_randomswitch(self)
        return []

    def __str__(self):
        ret = "random_switch({}, {}, {}, {}) {{\n".format(
            str(next(iter(self.feature_set))), str(self.type), str(self.name), str(self.triggers)
        )
        for dep in self.dependent:
            ret += "dependent: {};\n".format(dep)
        for indep in self.independent:
            ret += "independent: {};\n".format(indep)
        for choice in self.choices:
            ret += str(choice) + "\n"
        ret += "}\n"
        return ret


class RandomChoice:
    """
    Class to hold one of the possible choices in a random_switch

    @ivar probability: Relative chance for this choice to be chosen
    @type probability: L{Expression}

    @ivar result: Result of this choice, either another action2 or a return value
    @type result: L{SwitchValue}
    """

    def __init__(self, probability, result):
        self.probability = probability
        if result.value is None:
            raise generic.ScriptError(
                "Returning the computed value is not possible in a random_switch, as there is no computed value.",
                result.pos,
            )
        self.result = result

    def reduce_expressions(self, var_scope):
        self.probability = self.probability.reduce_constant(global_constants.const_list)
        if self.probability.value <= 0:
            raise generic.ScriptError("Random probability must be higher than 0", self.probability.pos)
        self.result.value = action2var.reduce_varaction2_expr(self.result.value, var_scope)

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Probability:")
        self.probability.debug_print(indentation + 2)

        generic.print_dbg(indentation, "Result:")
        self.result.debug_print(indentation + 2)

    def __str__(self):
        return "{}: {}".format(self.probability, self.result)



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic

from .base_expression import ConstantNumeric, Expression, Type


class TernaryOp(Expression):
    def __init__(self, guard, expr1, expr2, pos):
        Expression.__init__(self, pos)
        self.guard = guard
        self.expr1 = expr1
        self.expr2 = expr2

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Ternary operator")
        generic.print_dbg(indentation, "Guard:")
        self.guard.debug_print(indentation + 2)
        generic.print_dbg(indentation, "Expression 1:")
        self.expr1.debug_print(indentation + 2)
        generic.print_dbg(indentation, "Expression 2:")
        self.expr2.debug_print(indentation + 2)

    def reduce(self, id_dicts=None, unknown_id_fatal=True):
        guard = self.guard.reduce(id_dicts)
        expr1 = self.expr1.reduce(id_dicts)
        expr2 = self.expr2.reduce(id_dicts)
        if isinstance(guard, ConstantNumeric):
            if guard.value != 0:
                return expr1
            else:
                return expr2
        if guard.type() != Type.INTEGER or expr1.type() != Type.INTEGER or expr2.type() != Type.INTEGER:
            if guard.type() == Type.SPRITEGROUP_REF:
                raise generic.ProcCallSyntaxError(guard.name, guard.pos)
            if expr1.type() == Type.SPRITEGROUP_REF:
                raise generic.ProcCallSyntaxError(expr1.name, expr1.pos)
            if expr2.type() == Type.SPRITEGROUP_REF:
                raise generic.ProcCallSyntaxError(expr2.name, expr2.pos)
            raise generic.ScriptError("All parts of the ternary operator (?:) must be integers.", self.pos)
        return TernaryOp(guard, expr1, expr2, self.pos)

    def supported_by_action2(self, raise_error):
        return (
            self.guard.supported_by_action2(raise_error)
            and self.expr1.supported_by_action2(raise_error)
            and self.expr2.supported_by_action2(raise_error)
        )

    def supported_by_actionD(self, raise_error):
        return (
            self.guard.supported_by_actionD(raise_error)
            and self.expr1.supported_by_actionD(raise_error)
            and self.expr2.supported_by_actionD(raise_error)
        )

    def collect_references(self):
        return self.guard.collect_references() + self.expr1.collect_references() + self.expr2.collect_references()

    def is_read_only(self):
        return self.guard.is_read_only() and self.expr1.is_read_only() and self.expr2.is_read_only()

    def is_boolean(self):
        return self.expr1.is_boolean() and self.expr2.is_boolean()

    def __eq__(self, other):
        return (
            other is not None
            and isinstance(other, TernaryOp)
            and self.guard == other.guard
            and self.expr1 == other.expr1
            and self.expr2 == other.expr2
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self.guard, self.expr1, self.expr2))

    def __str__(self):
        return "({} ? {} : {})".format(str(self.guard), str(self.expr1), str(self.expr2))



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import generic


class SpriteContainer:
    """
    Base class for all AST Nodes that contain real sprites
    Note that this does not inherit from BaseStatement,
    to (amongst other things) avoid various multiple inheritance issues with Spritesets

    @ivar block_type: Type of block (e.g. 'spriteset' ,'replace')
    @type block_type: C{str}

    @ivar block_name: Block-specific name
    @type block_name: L{Identifier}, or C{None} if N/A

    @ivar sprite_data: Mapping of (zoom level, bit-depth) to (sprite list, default file)
    @type sprite_data: C{dict} that maps (C{tuple} of (C{int}, C{int}))
                       to (C{tuple} of (C{list} of (L{RealSprite}, L{RecolourSprite} or L{TemplateUsage}),
                                        L{StringLiteral} or C{None},
                                        L{Position}))
    """

    sprite_blocks = {}

    def __init__(self, block_type, block_name):
        self.block_type = block_type
        self.block_name = block_name
        self.sprite_data = {}
        if block_name is not None:
            if block_name.value in SpriteContainer.sprite_blocks:
                raise generic.ScriptError(
                    "Block with name '{}' is already defined.".format(block_name.value), block_name.pos
                )
            SpriteContainer.sprite_blocks[block_name.value] = self

    def add_sprite_data(self, sprite_list, default_file, pos, zoom_level=0, bit_depth=8, default_mask_file=None):
        assert zoom_level in range(0, 6)
        assert bit_depth in (8, 32)
        key = (zoom_level, bit_depth)
        if key in self.sprite_data:
            msg = (
                "Sprites are already defined for {} '{}' for this zoom "
                + "level / bit depth combination. This data will be overridden."
            )
            msg = msg.format(self.block_type, self.block_name.value)
            generic.print_warning(generic.Warning.GENERIC, msg, pos)
        self.sprite_data[key] = (sprite_list, default_file, default_mask_file, pos)

    def get_all_sprite_data(self):
        """
        Get all sprite data.
        Sorting makes sure that the order is consistent, and that the normal zoom, 8bpp sprites appear first.

        @return: List of 6-tuples (sprite_list, default_file, default_mask_file, position, zoom_level, bit_depth).
        @rtype:  C{list} of C{tuple} of (C{list} of (L{RealSprite},
                                         L{RecolourSprite} or L{TemplateUsage}),
                                         L{StringLiteral} or C{None},
                                         L{Position},
                                         C{int},
                                         C{int})
        """
        return [val + key for key, val in sorted(self.sprite_data.items())]

    @classmethod
    def resolve_sprite_block(cls, block_name):
        if block_name.value in cls.sprite_blocks:
            return cls.sprite_blocks[block_name.value]
        raise generic.ScriptError(
            "Undeclared block identifier '{}' encountered".format(block_name.value), block_name.pos
        )



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

import glob
import os
import re

from nml import generic


def utf8_get_size(char):
    if char < 128:
        return 1
    if char < 2048:
        return 2
    if char < 65536:
        return 3
    return 4


DEFAULT_LANGUAGE = 0x7F
DEFAULT_LANGNAME = "english.lng"


def validate_string(string):
    """
    Check if a given string refers to a string that is translated in the language
    files and raise an error otherwise.

    @param string: The string to validate.
    @type  string: L{expression.String}
    """
    if string.name.value not in default_lang.strings:
        raise generic.ScriptError('Unknown string "{}"'.format(string.name.value), string.pos)


def is_ascii_string(string):
    """
    Check whether a given string can be written using the ASCII codeset or
    that we need Unicode.

    @param string: The string to check.
    @type  string: C{str}

    @return: True iff the string is ascii-only.
    @rtype:  C{bool}
    """
    assert isinstance(string, str)
    i = 0
    while i < len(string):
        if string[i] != "\\":
            if ord(string[i]) >= 0x7B:
                return False
            i += 1
        else:
            if string[i + 1] in ("\\", '"'):
                i += 2
            elif string[i + 1] == "U":
                return False
            else:
                i += 3
    return True


def get_string_size(string, final_zero=True, force_ascii=False):
    """
    Get the size (in bytes) of a given string.

    @param string: The string to check.
    @type  string: C{str}

    @param final_zero: Whether to account for a zero-byte directly after the string.
    @type  final_zero: C{bool}

    @param force_ascii: When true, make sure the string is written as ascii as opposed to unicode.
    @type  force_ascii: C{bool}

    @return: The length (in bytes) of the given string.
    @rtype:  C{int}

    @raise generic.ScriptError: force_ascii and not is_ascii_string(string).
    """
    size = 0
    if final_zero:
        size += 1
    if not is_ascii_string(string):
        if force_ascii:
            raise generic.ScriptError("Expected ascii string but got a unicode string")
        size += 2
    i = 0
    while i < len(string):
        if string[i] != "\\":
            size += utf8_get_size(ord(string[i]))
            i += 1
        else:
            if string[i + 1] in ("\\", '"'):
                size += 1
                i += 2
            elif string[i + 1] == "U":
                size += utf8_get_size(int(string[i + 2 : i + 6], 16))
                i += 6
            else:
                size += 1
                i += 3
    return size


def get_translation(string, lang_id=DEFAULT_LANGUAGE):
    """
    Get the translation of a given string in a certain language. If there is no
    translation available in the given language return the default translation.

    @param string: the string to get the translation for.
    @type  string: L{expression.String}

    @param lang_id: The language id of the language to translate the string into.
    @type  lang_id: C{int}

    @return: Translation of the given string in the given language.
    @rtype:  C{str}
    """
    for lang_pair in langs:
        langid, lang = lang_pair
        if langid != lang_id:
            continue
        if string.name.value not in lang.strings:
            break
        return lang.get_string(string, lang_id)
    return default_lang.get_string(string, lang_id)


def get_translations(string):
    """
    Get a list of language ids that have a translation for the given string.

    @param string: the string to get the translations for.
    @type  string: L{expression.String}

    @return: List of languages that translate the given string.
    @rtype:  C{list} of C{int}
    """
    translations = []
    for lang_pair in langs:
        langid, lang = lang_pair
        assert langid is not None
        if string.name.value in lang.strings and lang.get_string(string, langid) != default_lang.get_string(
            string, langid
        ):
            translations.append(langid)

    # Also check for translated substrings
    import nml.expression

    for param in string.params:
        if not isinstance(param, nml.expression.String):
            continue
        param_translations = get_translations(param)
        translations.extend([langid for langid in param_translations if langid not in translations])

    return translations


def com_parse_comma(val, lang_id):
    val = val.reduce_constant()
    return str(val)


def com_parse_hex(val, lang_id):
    val = val.reduce_constant()
    return "0x{:X}".format(val.value)


def com_parse_string(val, lang_id):
    import nml.expression

    if not isinstance(val, (nml.expression.StringLiteral, nml.expression.String)):
        raise generic.ScriptError("Expected a (literal) string", val.pos)
    if isinstance(val, nml.expression.String):
        # Check that the string exists
        if val.name.value not in default_lang.strings:
            raise generic.ScriptError('Substring "{}" does not exist'.format(val.name.value), val.pos)
        return get_translation(val, lang_id)
    return val.value


# fmt: off
commands = {
    # Special characters / glyphs
    "":               {"unicode": r"\0D",       "ascii": r"\0D"},
    "{":              {"unicode": r"{",         "ascii": r"{"},
    "NBSP":           {"unicode": r"\U00A0"},  # character A0 is used as up arrow in TTD, so don"t use ASCII here.
    "COPYRIGHT":      {"unicode": r"\U00A9",    "ascii": r"\A9"},
    "TRAIN":          {"unicode": r"\UE0B4",    "ascii": r"\B4"},
    "LORRY":          {"unicode": r"\UE0B5",    "ascii": r"\B5"},
    "BUS":            {"unicode": r"\UE0B6",    "ascii": r"\B6"},
    "PLANE":          {"unicode": r"\UE0B7",    "ascii": r"\B7"},
    "SHIP":           {"unicode": r"\UE0B8",    "ascii": r"\B8"},

    # Change the font size.
    "TINYFONT":       {"unicode": r"\0E",       "ascii": r"\0E"},
    "BIGFONT":        {"unicode": r"\0F",       "ascii": r"\0F"},

    "COMMA":          {"unicode": r"\UE07B",    "ascii": r"\7B", "size": 4, "parse": com_parse_comma},
    "SIGNED_WORD":    {"unicode": r"\UE07C",    "ascii": r"\7C", "size": 2, "parse": com_parse_comma},
    "UNSIGNED_WORD":  {"unicode": r"\UE07E",    "ascii": r"\7E", "size": 2, "parse": com_parse_comma},
    "CURRENCY":       {"unicode": r"\UE07F",    "ascii": r"\7F", "size": 4},
    "STRING": {
        "unicode":    r"\UE080",
        "ascii":      r"\80",
        "allow_case": True,
        "size":       2,
        "parse":      com_parse_string
    },
    "DATE1920_LONG":  {"unicode": r"\UE082",    "ascii": r"\82", "size": 2},
    "DATE1920_SHORT": {"unicode": r"\UE083",    "ascii": r"\83", "size": 2},
    "VELOCITY":       {"unicode": r"\UE084",    "ascii": r"\84", "size": 2},
    "SKIP":           {"unicode": r"\UE085",    "ascii": r"\85", "size": 2},
    "VOLUME":         {"unicode": r"\UE087",    "ascii": r"\87", "size": 2},
    "HEX":            {"unicode": r"\UE09A\08", "ascii": r"\9A\08", "size": 4, "parse": com_parse_hex},
    "STATION":        {"unicode": r"\UE09A\0C", "ascii": r"\9A\0C", "size": 2},
    "WEIGHT":         {"unicode": r"\UE09A\0D", "ascii": r"\9A\0D", "size": 2},
    "DATE_LONG":      {"unicode": r"\UE09A\16", "ascii": r"\9A\16", "size": 4},
    "DATE_SHORT":     {"unicode": r"\UE09A\17", "ascii": r"\9A\17", "size": 4},
    "POWER":          {"unicode": r"\UE09A\18", "ascii": r"\9A\18", "size": 2},
    "VOLUME_SHORT":   {"unicode": r"\UE09A\19", "ascii": r"\9A\19", "size": 2},
    "WEIGHT_SHORT":   {"unicode": r"\UE09A\1A", "ascii": r"\9A\1A", "size": 2},
    "CARGO_LONG":     {"unicode": r"\UE09A\1B", "ascii": r"\9A\1B", "size": 2 * 2},
    "CARGO_SHORT":    {"unicode": r"\UE09A\1C", "ascii": r"\9A\1C", "size": 2 * 2},
    "CARGO_TINY":     {"unicode": r"\UE09A\1D", "ascii": r"\9A\1D", "size": 2 * 2},
    "CARGO_NAME":     {"unicode": r"\UE09A\1E", "ascii": r"\9A\1E", "size": 2},
    "FORCE":          {"unicode": r"\UE09A\21", "ascii": r"\9A\21", "size": 4},

    # Colors
    "BLUE":           {"unicode": r"\UE088",    "ascii": r"\88"},
    "SILVER":         {"unicode": r"\UE089",    "ascii": r"\89"},
    "GOLD":           {"unicode": r"\UE08A",    "ascii": r"\8A"},
    "RED":            {"unicode": r"\UE08B",    "ascii": r"\8B"},
    "PURPLE":         {"unicode": r"\UE08C",    "ascii": r"\8C"},
    "LTBROWN":        {"unicode": r"\UE08D",    "ascii": r"\8D"},
    "ORANGE":         {"unicode": r"\UE08E",    "ascii": r"\8E"},
    "GREEN":          {"unicode": r"\UE08F",    "ascii": r"\8F"},
    "YELLOW":         {"unicode": r"\UE090",    "ascii": r"\90"},
    "DKGREEN":        {"unicode": r"\UE091",    "ascii": r"\91"},
    "CREAM":          {"unicode": r"\UE092",    "ascii": r"\92"},
    "BROWN":          {"unicode": r"\UE093",    "ascii": r"\93"},
    "WHITE":          {"unicode": r"\UE094",    "ascii": r"\94"},
    "LTBLUE":         {"unicode": r"\UE095",    "ascii": r"\95"},
    "GRAY":           {"unicode": r"\UE096",    "ascii": r"\96"},
    "DKBLUE":         {"unicode": r"\UE097",    "ascii": r"\97"},
    "BLACK":          {"unicode": r"\UE098",    "ascii": r"\98"},

    "PUSH_COLOUR":    {"unicode": r"\UE09A\1F", "ascii": r"\9A\1F"},
    "POP_COLOUR":     {"unicode": r"\UE09A\20", "ascii": r"\9A\20"},

    # Deprecated string codes
    "DWORD_S":        {"unicode": r"\UE07B",    "ascii": r"\7B", "deprecated": True, "size": 4},
    "PARAM":          {"unicode": r"\UE07B",    "ascii": r"\7B", "deprecated": True, "size": 4},
    "WORD_S":         {"unicode": r"\UE07C",    "ascii": r"\7C", "deprecated": True, "size": 2},
    "BYTE_S":         {"unicode": r"\UE07D",    "ascii": r"\7D", "deprecated": True},
    "WORD_U":         {"unicode": r"\UE07E",    "ascii": r"\7E", "deprecated": True, "size": 2},
    "POP_WORD":       {"unicode": r"\UE085",    "ascii": r"\85", "deprecated": True, "size": 2},
    "CURRENCY_QWORD": {"unicode": r"\UE09A\01", "ascii": r"\9A\01", "deprecated": True},
    "PUSH_WORD":      {"unicode": r"\UE09A\03", "ascii": r"\9A\03", "deprecated": True},
    "UNPRINT":        {"unicode": r"\UE09A\04", "ascii": r"\9A\04", "deprecated": True},
    "BYTE_HEX":       {"unicode": r"\UE09A\06", "ascii": r"\9A\06", "deprecated": True},
    "WORD_HEX":       {"unicode": r"\UE09A\07", "ascii": r"\9A\07", "deprecated": True, "size": 2},
    "DWORD_HEX":      {"unicode": r"\UE09A\08", "ascii": r"\9A\08", "deprecated": True, "size": 4},
    "QWORD_HEX":      {"unicode": r"\UE09A\0B", "ascii": r"\9A\0B", "deprecated": True},
    "WORD_S_TONNES":  {"unicode": r"\UE09A\0D", "ascii": r"\9A\0D", "deprecated": True, "size": 2},
}
# fmt: on

special_commands = [
    "P",
    "G",
    "G=",
]


def read_extra_commands(custom_tags_file):
    """
    @param custom_tags_file: Filename of the custom tags file.
    @type  custom_tags_file: C{str}
    """
    if not os.access(custom_tags_file, os.R_OK):
        # Failed to open custom_tags.txt, ignore this
        return
    line_no = 0
    with open(generic.find_file(custom_tags_file), "r", encoding="utf-8") as fh:
        for line in fh:
            line_no += 1
            line = line.strip()
            if len(line) == 0 or line[0] == "#":
                continue

            i = line.find(":")
            if i == -1:
                raise generic.ScriptError("Line has no ':' delimiter.", generic.LinePosition(custom_tags_file, line_no))
            name = line[:i].strip()
            value = line[i + 1 :]
            if name in commands:
                generic.print_warning(
                    generic.Warning.GENERIC,
                    'Overwriting existing tag "' + name + '".',
                    generic.LinePosition(custom_tags_file, line_no),
                )
            commands[name] = {"unicode": value}
            if is_ascii_string(value):
                commands[name]["ascii"] = value


class StringCommand:
    """
    Instantiated string command.

    @ivar name: Name of the string command.
    @type name: C{str}

    @ivar case: ???
    @type case: ???

    @ivar arguments: Arguments of the instantiated command.
    @type arguments: C{list} of ???

    @ivar offset: Index to an argument of the P or G string command (ie "2" in {P 2 ...}), if available.
    @type offset: C{int} or C{None}

    @ivar str_pos: String argument index ("2" in {2:FOO..}, if available.
    @type str_pos: C{int} or C{None}

    @ivar pos: Position of the string.
    @type pos: L{Position}
    """

    def __init__(self, name, str_pos, pos):
        assert name in commands or name in special_commands
        self.name = name
        self.case = None
        self.arguments = []
        self.offset = None
        self.str_pos = str_pos
        self.pos = pos

    def set_arguments(self, arg_string):
        start = -1
        cur = 0
        quoted = False
        whitespace = " \t"
        while cur < len(arg_string):
            if start != -1:
                if (quoted and arg_string[cur] == '"') or (not quoted and arg_string[cur] in whitespace):
                    if (
                        not quoted
                        and self.offset is None
                        and len(self.arguments) == 0
                        and isint(arg_string[start:cur])
                        and self.name in ("P", "G")
                    ):
                        self.offset = int(arg_string[start:cur])
                    else:
                        self.arguments.append(arg_string[start:cur])
                    start = -1
            elif arg_string[cur] not in whitespace:
                quoted = arg_string[cur] == '"'
                start = cur + 1 if quoted else cur
            cur += 1
        if start != -1 and not quoted:
            self.arguments.append(arg_string[start:])
            start = -1
        return start == -1

    def validate_arguments(self, lang):
        if lang.langid == DEFAULT_LANGUAGE:
            return
        if self.name == "P":
            if not lang.has_plural_pragma():
                raise generic.ScriptError("Using {P} without a ##plural pragma", self.pos)
            if len(self.arguments) != lang.get_num_plurals():
                raise generic.ScriptError(
                    "Invalid number of arguments to plural command, expected {:d} but got {:d}".format(
                        lang.get_num_plurals(), len(self.arguments)
                    ),
                    self.pos,
                )
        elif self.name == "G":
            if not lang.has_gender_pragma():
                raise generic.ScriptError("Using {G} without a ##gender pragma", self.pos)
            if len(self.arguments) != len(lang.genders):
                raise generic.ScriptError(
                    "Invalid number of arguments to gender command, expected {:d} but got {:d}".format(
                        len(lang.genders), len(self.arguments)
                    ),
                    self.pos,
                )
        elif self.name == "G=":
            if not lang.has_gender_pragma():
                raise generic.ScriptError("Using {G=} without a ##gender pragma", self.pos)
            if len(self.arguments) != 1:
                raise generic.ScriptError(
                    "Invalid number of arguments to set-gender command, expected {:d} but got {:d}".format(
                        1, len(self.arguments)
                    ),
                    self.pos,
                )
        elif len(self.arguments) != 0:
            raise generic.ScriptError('Unexpected arguments to command "{}"'.format(self.name), self.pos)

    def parse_string(self, str_type, lang, wanted_lang_id, prev_command, stack, static_args):
        """
        Convert the string command to output text.

        @param str_type: Expected type of result text, C{"unicode"} or C{"ascii"}.
        @type  str_type: C{str}

        @param lang: Language of the string.
        @type  lang: L{Language}

        @param wanted_lang_id: Language-id to use for interpreting the command
                               (this string may be from another language, e.g. with missing strings).

        @param prev_command: Argument of previous string command (parameter number, size).
        @type  prev_command: C{tuple} or C{None}

        @param stack: Stack of available arguments (list of (parameter number, size)).
        @type  stack: C{list} of C{tuple} (C{int}, C{int}) or C{None}

        @param static_args: Static command arguments.
        """
        if self.name in commands:
            if not self.is_important_command():
                return commands[self.name][str_type]
            # Compute position of the argument in the stack.
            stack_pos = 0
            for pos, size in stack:
                if pos == self.str_pos:
                    break
                stack_pos += size
            self_size = commands[self.name]["size"]
            stack.remove((self.str_pos, self_size))
            if self.str_pos < len(static_args):
                if "parse" not in commands[self.name]:
                    raise generic.ScriptError(
                        "Provided a static argument for string command '{}' which is invalid".format(self.name),
                        self.pos,
                    )
                # Parse commands using the wanted (not current) lang id, so translations are used if present
                return commands[self.name]["parse"](static_args[self.str_pos], wanted_lang_id)
            prefix = ""
            suffix = ""
            if self.case:
                prefix += STRING_SELECT_CASE[str_type] + "\\{:02X}".format(self.case)
            if stack_pos + self_size > 8:
                raise generic.ScriptError(
                    "Trying to read an argument from the stack without reading the arguments before", self.pos
                )
            if self_size == 4 and stack_pos == 4:
                prefix += STRING_ROTATE[str_type] + STRING_ROTATE[str_type]
            elif self_size == 4 and stack_pos == 2:
                prefix += STRING_PUSH_WORD[str_type] + STRING_ROTATE[str_type] + STRING_ROTATE[str_type]
                suffix += STRING_SKIP[str_type]
            elif self_size == 2 and stack_pos == 6:
                prefix += STRING_ROTATE[str_type]
            elif self_size == 2 and stack_pos == 4:
                prefix += STRING_PUSH_WORD[str_type] + STRING_ROTATE[str_type]
                suffix += STRING_SKIP[str_type]
            elif self_size == 2 and stack_pos == 2:
                prefix += STRING_PUSH_WORD[str_type] + STRING_PUSH_WORD[str_type] + STRING_ROTATE[str_type]
                suffix += STRING_SKIP[str_type] + STRING_SKIP[str_type]
            else:
                assert stack_pos == 0
            return prefix + commands[self.name][str_type] + suffix
        assert self.name in special_commands
        # Create a local copy because we shouldn't modify the original
        offset = self.offset
        if offset is None:
            if self.name == "P":
                if not prev_command:
                    raise generic.ScriptError(
                        "A plural choice list {P} has to be preceded by another string code or provide an offset",
                        self.pos,
                    )
                offset = prev_command[0]
            else:
                if not stack:
                    raise generic.ScriptError(
                        "A gender choice list {G} has to be followed by another string code or provide an offset",
                        self.pos,
                    )
                offset = stack[0][0]
        offset -= len(static_args)
        if self.name == "P":
            if offset < 0:
                return self.arguments[lang.static_plural_form(static_args[offset]) - 1]
            ret = BEGIN_PLURAL_CHOICE_LIST[str_type] + "\\{:02X}".format(0x80 + offset)
            for idx, arg in enumerate(self.arguments):
                if idx == len(self.arguments) - 1:
                    ret += CHOICE_LIST_DEFAULT[str_type]
                else:
                    ret += CHOICE_LIST_ITEM[str_type] + "\\{:02X}".format(idx + 1)
                ret += arg
            ret += CHOICE_LIST_END[str_type]
            return ret
        if self.name == "G":
            if offset < 0:
                return self.arguments[lang.static_gender(static_args[offset]) - 1]
            ret = BEGIN_GENDER_CHOICE_LIST[str_type] + "\\{:02X}".format(0x80 + offset)
            for idx, arg in enumerate(self.arguments):
                if idx == len(self.arguments) - 1:
                    ret += CHOICE_LIST_DEFAULT[str_type]
                else:
                    ret += CHOICE_LIST_ITEM[str_type] + "\\{:02X}".format(idx + 1)
                ret += arg
            ret += CHOICE_LIST_END[str_type]
            return ret
        # Not reached
        raise ValueError("Unexpected string command '{}'".format(self.name))

    def get_type(self):
        if self.name in commands:
            if "ascii" in commands[self.name]:
                return "ascii"
            else:
                return "unicode"
        if self.name == "P" or self.name == "G":
            for arg in self.arguments:
                if not is_ascii_string(arg):
                    return "unicode"
        return "ascii"

    def is_important_command(self):
        if self.name in special_commands:
            return False
        return "size" in commands[self.name]

    def get_arg_size(self):
        return commands[self.name]["size"]


# Characters that are valid in hex numbers
VALID_HEX = "0123456789abcdefABCDEF"


def is_valid_hex(string):
    return all(c in VALID_HEX for c in string)


def validate_escapes(string, pos):
    """
    Validate that all escapes (starting with a backslash) are correct.
    When an invalid escape is encountered, an error is thrown

    @param string: String to validate
    @type string: C{str}

    @param pos: Position information
    @type pos: L{Position}
    """
    i = 0
    while i < len(string):
        # find next '\'
        i = string.find("\\", i)
        if i == -1:
            break

        if i + 1 >= len(string):
            raise generic.ScriptError("Unexpected end-of-line encountered after '\\'", pos)
        if string[i + 1] in ("\\", '"'):
            i += 2
        elif string[i + 1] == "U":
            if i + 5 >= len(string) or not is_valid_hex(string[i + 2 : i + 6]):
                raise generic.ScriptError("Expected 4 hexadecimal characters after '\\U'", pos)
            i += 6
        else:
            if i + 2 >= len(string) or not is_valid_hex(string[i + 1 : i + 3]):
                raise generic.ScriptError("Expected 2 hexadecimal characters after '\\'", pos)
            i += 3


class NewGRFString:
    """
    A string text in a language.

    @ivar string: String text.
    @type string: C{str}

    @ivar cases: Mapping of cases to ...
    @type cases: ???

    @ivar components: Split string text.
    @type components: C{list} of (C{str} or L{StringCommand})

    @ivar pos: Position of the string text.
    @type pos: L{Position}
    """

    def __init__(self, string, lang, pos):
        """
        Construct a L{NewGRFString}, and break down the string text into text literals and string commands.

        @param string: String text.
        @type  string: C{str}

        @param lang: Language containing this string.
        @type  lang: L{Language}

        @param pos: Position of the string text.
        @type  pos: L{Position}
        """
        validate_escapes(string, pos)
        self.string = string
        self.cases = {}
        self.components = []
        self.pos = pos
        idx = 0
        while idx < len(string):
            if string[idx] != "{":
                j = string.find("{", idx)
                if j == -1:
                    self.components.append(string[idx:])
                    break
                self.components.append(string[idx:j])
                idx = j

            start = idx + 1
            end = start
            cmd_pos = None
            if start >= len(string):
                raise generic.ScriptError("Expected '}' before end-of-line.", pos)
            if string[start].isdigit():
                while end < len(string) and string[end].isdigit():
                    end += 1
                if end == len(string) or string[end] != ":":
                    raise generic.ScriptError("Error while parsing position part of string command", pos)
                cmd_pos = int(string[start:end])
                start = end + 1
                end = start
            # Read the command name
            while end < len(string) and string[end] not in "} =.":
                end += 1
            command_name = string[start:end]
            if end < len(string) and string[end] == "=":
                command_name += "="
            if command_name not in commands and command_name not in special_commands:
                raise generic.ScriptError('Undefined command "{}"'.format(command_name), pos)
            if command_name in commands and "deprecated" in commands[command_name]:
                generic.print_warning(
                    generic.Warning.DEPRECATION,
                    "String code '{}' has been deprecated and will be removed soon".format(command_name),
                    pos,
                )
                del commands[command_name]["deprecated"]
            #
            command = StringCommand(command_name, cmd_pos, pos)
            if end >= len(string):
                raise generic.ScriptError("Missing '}}' from command \"{}\"".format(string[start:]), pos)
            if string[end] == ".":
                if command_name not in commands or "allow_case" not in commands[command_name]:
                    raise generic.ScriptError('Command "{}" can\'t have a case'.format(command_name), pos)
                case_start = end + 1
                end = case_start
                while end < len(string) and string[end] not in "} ":
                    end += 1
                case = string[case_start:end]
                if lang.cases is None or case not in lang.cases:
                    raise generic.ScriptError('Invalid case-name "{}"'.format(case), pos)
                command.case = lang.cases[case]
            if string[end] != "}":
                arg_start = end + 1
                end = string.find("}", end + 1)
                if end == -1 or not command.set_arguments(string[arg_start:end]):
                    raise generic.ScriptError("Missing '}}' from command \"{}\"".format(string[start:]), pos)
            command.validate_arguments(lang)
            if command_name == "G=" and self.components:
                raise generic.ScriptError("Set-gender command {G=} must be at the start of the string", pos)
            self.components.append(command)
            idx = end + 1

        if (
            len(self.components) > 0
            and isinstance(self.components[0], StringCommand)
            and self.components[0].name == "G="
        ):
            self.gender = self.components[0].arguments[0]
            if self.gender not in lang.genders:
                raise generic.ScriptError("Invalid gender name '{}'".format(self.gender), pos)
            self.components.pop(0)
        else:
            self.gender = None
        cmd_pos = 0
        for cmd in self.components:
            if not (isinstance(cmd, StringCommand) and cmd.is_important_command()):
                continue
            if cmd.str_pos is None:
                cmd.str_pos = cmd_pos
            cmd_pos = cmd.str_pos + 1

    def get_type(self):
        """
        Get the type of string text.

        @return: C{"unicode"} if Unicode is required for expressing the string text, else C{"ascii"}.
        @rtype:  C{str}
        """
        for comp in self.components:
            if isinstance(comp, StringCommand):
                if comp.get_type() == "unicode":
                    return "unicode"
            else:
                if not is_ascii_string(comp):
                    return "unicode"
        for case in self.cases.values():
            if case.get_type() == "unicode":
                return "unicode"
        return "ascii"

    def remove_non_default_commands(self):
        i = 0
        while i < len(self.components):
            comp = self.components[i]
            if isinstance(comp, StringCommand):
                if comp.name == "P" or comp.name == "G":
                    self.components[i] = comp.arguments[-1] if comp.arguments else ""
            i += 1

    def parse_string(self, str_type, lang, wanted_lang_id, static_args):
        """
        Convert the string text to output text.
        """
        ret = ""
        stack = [(idx, size) for idx, size in enumerate(self.get_command_sizes())]
        prev_command = None
        for comp in self.components:
            if isinstance(comp, StringCommand):
                next_command = stack[0] if stack else None
                ret += comp.parse_string(str_type, lang, wanted_lang_id, prev_command, stack, static_args)

                if (comp.name in commands) and comp.is_important_command():
                    prev_command = next_command
            else:
                ret += comp
        return ret

    def get_command_sizes(self):
        """
        Get the size of each string parameter.
        """
        sizes = {}
        for cmd in self.components:
            if not (isinstance(cmd, StringCommand) and cmd.is_important_command()):
                continue
            if cmd.str_pos in sizes:
                raise generic.ScriptError("Two or more string commands are using the same argument", self.pos)
            sizes[cmd.str_pos] = cmd.get_arg_size()
        sizes_list = []
        for idx in range(len(sizes)):
            if idx not in sizes:
                raise generic.ScriptError("String argument {:d} is not used".format(idx), self.pos)
            sizes_list.append(sizes[idx])
        return sizes_list

    def match_commands(self, other_string):
        return self.get_command_sizes() == other_string.get_command_sizes()


def isint(x, base=10):
    try:
        int(x, base)
        return True
    except ValueError:
        return False


NUM_PLURAL_FORMS = 14

CHOICE_LIST_ITEM = {"unicode": r"\UE09A\10", "ascii": r"\9A\10"}
CHOICE_LIST_DEFAULT = {"unicode": r"\UE09A\11", "ascii": r"\9A\11"}
CHOICE_LIST_END = {"unicode": r"\UE09A\12", "ascii": r"\9A\12"}
BEGIN_GENDER_CHOICE_LIST = {"unicode": r"\UE09A\13", "ascii": r"\9A\13"}
BEGIN_CASE_CHOICE_LIST = {"unicode": r"\UE09A\14", "ascii": r"\9A\14"}
BEGIN_PLURAL_CHOICE_LIST = {"unicode": r"\UE09A\15", "ascii": r"\9A\15"}
SET_STRING_GENDER = {"unicode": r"\UE09A\0E", "ascii": r"\9A\0E"}
STRING_SKIP = {"unicode": r"\UE085", "ascii": r"\85"}
STRING_ROTATE = {"unicode": r"\UE086", "ascii": r"\86"}
STRING_PUSH_WORD = {"unicode": r"\UE09A\03\20\20", "ascii": r"\9A\03\20\20"}
STRING_SELECT_CASE = {"unicode": r"\UE09A\0F", "ascii": r"\9A\0F"}


class Language:
    """
    @ivar default: Whether the language is the default language.
    @type default: C{bool}

    @ivar langid: Language id of the language, if known.
    @type langid: C{None} or C{int}

    @ivar plural: Plural type.
    @type plural: C{None} or C{int}

    @ivar genders:
    @type genders:

    @ivar gender_map:
    @type gender_map:

    @ivar cases:
    @type cases:

    @ivar case_map:
    @type case_map:

    @ivar strings: Language strings of the file.
    @type strings: C{dict} of
    """

    used_strings = []

    def __init__(self, default):
        self.default = default
        self.langid = None
        self.plural = None
        self.genders = None
        self.gender_map = {}
        self.cases = None
        self.case_map = {}
        self.strings = {}

    def get_num_plurals(self):
        if self.plural is None:
            return 0
        num_plurals = {
            0: 2,
            1: 1,
            2: 2,
            3: 3,
            4: 5,
            5: 3,
            6: 3,
            7: 3,
            8: 4,
            9: 2,
            10: 3,
            11: 2,
            12: 4,
            13: 4,
            14: 3,
        }
        return num_plurals[self.plural]

    def has_plural_pragma(self):
        return self.plural is not None

    def has_gender_pragma(self):
        return self.genders is not None

    def static_gender(self, expr):
        import nml.expression

        if isinstance(expr, nml.expression.StringLiteral):
            return len(self.genders)
        if not isinstance(expr, nml.expression.String):
            raise generic.ScriptError("{G} can only refer to a string argument")
        parsed = self.get_string(expr, self.langid)
        if parsed.find(SET_STRING_GENDER["ascii"]) == 0:
            return int(parsed[len(SET_STRING_GENDER["ascii"]) + 1 : len(SET_STRING_GENDER["ascii"]) + 3], 16)
        if parsed.find(SET_STRING_GENDER["unicode"]) == 0:
            return int(parsed[len(SET_STRING_GENDER["unicode"]) + 1 : len(SET_STRING_GENDER["unicode"]) + 3], 16)
        return len(self.genders)

    def static_plural_form(self, expr):
        # Return values are the same as "Plural index" here:
        # http://newgrf-specs.tt-wiki.net/wiki/StringCodes#Using_plural_forms
        val = expr.reduce_constant().value
        if self.plural == 0:
            return 1 if val == 1 else 2
        if self.plural == 1:
            return 1
        if self.plural == 2:
            return 1 if val in (0, 1) else 2
        if self.plural == 3:
            if val % 10 == 1 and val % 100 != 11:
                return 1
            return 2 if val == 0 else 3
        if self.plural == 4:
            if val == 1:
                return 1
            if val == 2:
                return 2
            if 3 <= val <= 6:
                return 3
            if 7 <= val <= 10:
                return 4
            return 5
        if self.plural == 5:
            if val % 10 == 1 and val % 100 != 11:
                return 1
            if 2 <= (val % 10) <= 9 and not 12 <= (val % 100) <= 19:
                return 2
            return 3
        if self.plural == 6:
            if val % 10 == 1 and val % 100 != 11:
                return 1
            if 2 <= (val % 10) <= 4 and not 12 <= (val % 100) <= 14:
                return 2
            return 3
        if self.plural == 7:
            if val == 0:
                return 1
            if 2 <= (val % 10) <= 4 and not 12 <= (val % 100) <= 14:
                return 2
            return 3
        if self.plural == 8:
            if val % 100 == 1:
                return 1
            if val % 100 == 2:
                return 2
            if val % 100 in (3, 4):
                return 3
            return 4
        if self.plural == 9:
            if val % 10 == 1 and val % 100 != 11:
                return 1
            return 2
        if self.plural == 10:
            if val == 1:
                return 1
            if 2 <= val <= 4:
                return 2
            return 3
        if self.plural == 11:
            if val % 10 in (0, 1, 3, 6, 7, 8):
                return 1
            return 2
        if self.plural == 12:
            if val == 1:
                return 1
            if val == 0 or 2 <= (val % 100) <= 10:
                return 2
            if 11 <= (val % 100) <= 19:
                return 3
            return 4
        if self.plural == 13:
            if val == 1 or val == 11:
                return 1
            if val == 2 or val == 12:
                return 2
            if (val >= 3 and val <= 10) or (val >= 13 and val <= 19):
                return 3
            return 4
        if self.plural == 14:
            if val == 1:
                return 1
            if val == 0 or 1 <= (val % 100) <= 19:
                return 2
            return 3
        raise AssertionError("Unknown plural type")

    def get_string(self, string, lang_id):
        """
        Lookup up a string by name/params and return the actual created string

        @param string: String object
        @type string: L{expression.String}

        @param lang_id: Language ID we are actually looking for.
                This may differ from the ID of this language,
                if the string is missing from the target language.
        @type lang_id: C{int}

        @return: The created string
        @rtype: C{str}
        """
        string_id = string.name.value
        assert isinstance(string_id, str)
        assert string_id in self.strings
        assert lang_id == self.langid or self.langid == DEFAULT_LANGUAGE

        if string_id not in Language.used_strings:
            Language.used_strings.append(string_id)

        str_type = self.strings[string_id].get_type()
        parsed_string = ""
        if self.strings[string_id].gender is not None:
            parsed_string += SET_STRING_GENDER[str_type] + "\\{:02X}".format(
                self.genders[self.strings[string_id].gender]
            )
        if len(self.strings[string_id].cases) > 0:
            parsed_string += BEGIN_CASE_CHOICE_LIST[str_type]
            for case_name, case_string in sorted(self.strings[string_id].cases.items()):
                case_id = self.cases[case_name]
                parsed_string += (
                    CHOICE_LIST_ITEM[str_type]
                    + "\\{:02X}".format(case_id)
                    + case_string.parse_string(str_type, self, lang_id, string.params)
                )
            parsed_string += CHOICE_LIST_DEFAULT[str_type]
        parsed_string += self.strings[string_id].parse_string(str_type, self, lang_id, string.params)
        if len(self.strings[string_id].cases) > 0:
            parsed_string += CHOICE_LIST_END[str_type]
        return parsed_string

    def handle_grflangid(self, data, pos):
        """
        Handle a 'grflangid' pragma.

        @param data: Data of the pragma.
        @type  data: C{str}
        """
        if self.langid is not None:
            raise generic.ScriptError("grflangid already set", pos)
        lang_text = data.strip()
        try:
            value = int(lang_text, 16)
        except ValueError:
            raise generic.ScriptError("Invalid grflangid {!r}".format(lang_text), pos)
        if value < 0 or value >= 0x7F:
            raise generic.ScriptError("Invalid grflangid", pos)
        self.langid = value

    def handle_plural(self, data, pos):
        """
        Handle a 'plural' pragma.

        @param data: Data of the pragma.
        @type  data: C{str}
        """
        if self.plural is not None:
            raise generic.ScriptError("plural form already set", pos)
        try:
            # Explicitly force base 10 like in OpenTTD's lang file
            value = int(data, 10)
        except ValueError:
            raise generic.ScriptError("Invalid plural form", pos)
        if value < 0 or value > NUM_PLURAL_FORMS:
            raise generic.ScriptError("Invalid plural form", pos)
        self.plural = value

    def handle_gender(self, data, pos):
        """
        Handle a 'gender' pragma.

        @param data: Data of the pragma.
        @type  data: C{str}
        """
        if self.genders is not None:
            raise generic.ScriptError("Genders already defined", pos)
        self.genders = {}
        for idx, gender in enumerate(data.split()):
            self.genders[gender] = idx + 1
            self.gender_map[gender] = []

    def handle_map_gender(self, data, pos):
        """
        Handle a 'map_gender' pragma.

        @param data: Data of the pragma.
        @type  data: C{str}
        """
        if self.genders is None:
            raise generic.ScriptError("##map_gender is not allowed before ##gender", pos)
        genders = data.split()
        if len(genders) != 2:
            raise generic.ScriptError("Invalid ##map_gender line", pos)
        if genders[0] not in self.genders:
            raise generic.ScriptError("Trying to map non-existing gender '{}'".format(genders[0]), pos)
        self.gender_map[genders[0]].append(genders[1])

    def handle_case(self, data, pos):
        """
        Handle a 'case' pragma.

        @param data: Data of the pragma.
        @type  data: C{str}
        """
        if self.cases is not None:
            raise generic.ScriptError("Cases already defined", pos)
        self.cases = {}
        for idx, case in enumerate(data.split()):
            self.cases[case] = idx + 1
            self.case_map[case] = []

    def handle_map_case(self, data, pos):
        """
        Handle a 'map_case' pragma.

        @param data: Data of the pragma.
        @type  data: C{str}
        """
        if self.cases is None:
            raise generic.ScriptError("##map_case is not allowed before ##case", pos)
        cases = data.split()
        if len(cases) != 2:
            raise generic.ScriptError("Invalid ##map_case line", pos)
        if cases[0] not in self.cases:
            raise generic.ScriptError("Trying to map non-existing case '{}'".format(cases[0]), pos)
        self.case_map[cases[0]].append(cases[1])

    def handle_text(self, string, case, value, pos):
        """
        Handle a text string definition.

        @param string: Name of the string.
        @type  string: C{str}

        @param case: The case being defined (optional).
        @type  case: C{str} or C{None}

        @param value: Value of the string.
        @type  value: C{str}
        """
        if not re.match("[A-Za-z_0-9]+$", string):
            raise generic.ScriptError('Invalid string name "{}"'.format(string), pos)

        if string in self.strings and case is None:
            raise generic.ScriptError('String name "{}" is used multiple times'.format(string), pos)

        if self.default:
            self.strings[string] = NewGRFString(value, self, pos)
            self.strings[string].remove_non_default_commands()
        else:
            if string not in default_lang.strings:
                generic.print_warning(
                    generic.Warning.GENERIC, 'String name "{}" does not exist in master file'.format(string), pos
                )
                return

            newgrf_string = NewGRFString(value, self, pos)
            if not default_lang.strings[string].match_commands(newgrf_string):
                generic.print_warning(
                    generic.Warning.GENERIC,
                    'String commands don\'t match with master file "{}"'.format(DEFAULT_LANGNAME),
                    pos,
                )
                return

            if case is None:
                self.strings[string] = newgrf_string
            else:
                if string not in self.strings:
                    generic.print_warning(generic.Warning.GENERIC, "String with case used before the base string", pos)
                    return
                if self.cases is None or case not in self.cases:
                    generic.print_warning(generic.Warning.GENERIC, 'Invalid case name "{}"'.format(case), pos)
                    return
                if case in self.strings[string].cases:
                    raise generic.ScriptError('String name "{}.{}" is used multiple times'.format(string, case), pos)
                if newgrf_string.gender:
                    generic.print_warning(
                        generic.Warning.GENERIC, "Case-strings can't set the gender, only the base string can", pos
                    )
                    return
                self.strings[string].cases[case] = newgrf_string

    def handle_string(self, line, pos):
        if len(line) == 0:
            # Silently ignore empty lines.
            return

        elif len(line) > 2 and line[:2] == "##" and not line[:3] == "###":
            # "##pragma" line, call relevant handler.
            if self.default:
                # Default language ignores all pragmas.
                return
            pragmas = {
                "##grflangid": self.handle_grflangid,
                "##plural": self.handle_plural,
                "##gender": self.handle_gender,
                "##map_gender": self.handle_map_gender,
                "##map_case": self.handle_map_case,
                "##case": self.handle_case,
            }
            try:
                pragma, value = line.split(" ", maxsplit=1)
                handler = pragmas[pragma]
            except (ValueError, KeyError):
                raise generic.ScriptError("Invalid pragma", pos)
            handler(value, pos)

        elif line[0] == "#":
            # Normal comment, ignore
            return

        else:
            # Must be a line defining a string.
            try:
                name, value = line.split(":", maxsplit=1)
                name = name.rstrip()
            except ValueError:
                raise generic.ScriptError("Line has no ':' delimiter", pos)

            if "." in name:
                name, case = name.rsplit(".", maxsplit=1)
                if not case:
                    raise generic.ScriptError("No case after '.' delimiter", pos)
            else:
                case = None

            if self.default and case is not None:
                # Ignore cases for the default language
                return

            self.handle_text(name, case, value, pos)


default_lang = Language(True)
default_lang.langid = DEFAULT_LANGUAGE
langs = []


def parse_file(filename, default):
    """
    Read and parse a single language file.

    @param filename: The filename of the file to parse.
    @type  filename: C{str}

    @param default: True iff this is the default language.
    @type  default: C{bool}
    """
    lang = Language(False)
    try:
        with open(generic.find_file(filename), "r", encoding="utf-8") as fh:
            for idx, line in enumerate(fh):
                pos = generic.LinePosition(filename, idx + 1)
                line = line.rstrip("\n\r").lstrip("\ufeff")
                # The default language is processed twice here. Once as fallback langauge
                # and once as normal language.
                if default:
                    default_lang.handle_string(line, pos)
                lang.handle_string(line, pos)
    except UnicodeDecodeError:
        pos = generic.LanguageFilePosition(filename)
        if default:
            raise generic.ScriptError("The default language file contains non-utf8 characters.", pos)
        generic.print_warning(
            generic.Warning.GENERIC, "Language file contains non-utf8 characters. Ignoring (part of) the contents.", pos
        )
    except generic.ScriptError as err:
        if default:
            raise
        generic.print_warning(generic.Warning.GENERIC, err.value, err.pos)
    else:
        if lang.langid is None:
            generic.print_warning(
                generic.Warning.GENERIC,
                "Language file does not contain a ##grflangid pragma",
                generic.LanguageFilePosition(filename),
            )
        else:
            for lng in langs:
                if lng[0] == lang.langid:
                    msg = "Language file has the same ##grflangid (with number {:d}) as another language file".format(
                        lang.langid
                    )
                    raise generic.ScriptError(msg, generic.LanguageFilePosition(filename))
            langs.append((lang.langid, lang))


def read_lang_files(lang_dir, default_lang_file):
    """
    Read the language files containing the translations for string constants
    used in the NML specification.

    @param lang_dir: Name of the directory containing the language files.
    @type  lang_dir: C{str}

    @param default_lang_file: Filename of the language file that has the
                              default translation which will be used as
                              fallback for other languages.
    @type  default_lang_file: C{str}
    """
    global DEFAULT_LANGNAME

    DEFAULT_LANGNAME = default_lang_file
    if not os.path.exists(lang_dir + os.sep + default_lang_file):
        generic.print_warning(
            generic.Warning.GENERIC,
            'Default language file "{}" doesn\'t exist'.format(os.path.join(lang_dir, default_lang_file)),
        )
        return
    parse_file(lang_dir + os.sep + default_lang_file, True)
    for filename in glob.glob(lang_dir + os.sep + "*.lng"):
        if filename.endswith(default_lang_file):
            continue
        parse_file(filename, False)
    langs.sort()


def list_unused_strings():
    for string in default_lang.strings:
        if string in Language.used_strings:
            continue
        generic.print_warning(generic.Warning.GENERIC, 'String "{}" is unused'.format(string))



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic, global_constants
from nml.actions import action0, action2, action2var, action3
from nml.ast import base_statement, general

item_feature = None
item_id = None
item_size = None


class Item(base_statement.BaseStatementList):
    """
    AST-node representing an item block
    @ivar feature: Feature of the item
    @type feature: L{ConstantNumeric}

    @ivar name: Name of the item
    @type name: L{Identifier} or C{None} if N/A.

    @ivar id: Numeric ID of the item
    @type id: L{ConstantNumeric} or C{None} if N/A

    @ivar size: Size, used by houses only
    @type size: L{ConstantNumeric} or C{None} if N/A
    """

    def __init__(self, params, body, pos):
        base_statement.BaseStatementList.__init__(
            self, "item-block", pos, base_statement.BaseStatementList.LIST_TYPE_ITEM, body
        )

        if not (1 <= len(params) <= 4):
            raise generic.ScriptError(
                "Item block requires between 1 and 4 parameters, found {:d}.".format(len(params)), self.pos
            )
        self.feature = general.parse_feature(params[0])
        if self.feature.value in (0x08, 0x0C, 0x0E):
            raise generic.ScriptError("Defining item blocks for this feature is not allowed.", self.pos)
        self.name = params[1] if len(params) >= 2 else None
        self.id = params[2] if len(params) >= 3 else None

        if len(params) >= 4:
            if self.feature.value != 0x07:
                raise generic.ScriptError("item-block parameter 4 'size' may only be set for houses", params[3].pos)
            self.size = params[3].reduce_constant(global_constants.const_list)
            if self.size.value not in action0.house_sizes:
                raise generic.ScriptError("item-block parameter 4 'size' does not have a valid value", self.size.pos)
        else:
            self.size = None

    def register_names(self):
        if self.id is not None:
            self.id = self.id.reduce_constant(global_constants.const_list)
            if isinstance(self.id, expression.ConstantNumeric) and self.id.value == -1:
                self.id = None  # id == -1 means default

        if self.name:
            if not isinstance(self.name, expression.Identifier):
                raise generic.ScriptError("Item parameter 2 'name' should be an identifier", self.pos)
            if self.name.value in global_constants.item_names:
                existing_id = global_constants.item_names[self.name.value].id
                if self.id is not None and existing_id.value != self.id.value:
                    raise generic.ScriptError(
                        (
                            "Duplicate item with name '{}'."
                            " This item has already been assigned to id {:d}, cannot reassign to id {:d}"
                        ).format(self.name.value, existing_id.value, self.id.value),
                        self.pos,
                    )
                self.id = existing_id

        # We may have to reserve multiple item IDs for houses
        num_ids = action0.house_sizes[self.size.value] if self.size is not None else 1
        if self.id is None:
            self.id = expression.ConstantNumeric(action0.get_free_id(self.feature.value, num_ids, self.pos))
        elif not action0.check_id_range(self.feature.value, self.id.value, num_ids, self.id.pos):
            action0.mark_id_used(self.feature.value, self.id.value, num_ids)
        if self.name is not None:
            global_constants.item_names[self.name.value] = self
        base_statement.BaseStatementList.register_names(self)

    def pre_process(self):
        global item_feature, item_id, item_size
        item_id = self.id
        item_feature = self.feature.value
        item_size = self.size
        base_statement.BaseStatementList.pre_process(self)

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Item, feature", hex(self.feature.value))
        base_statement.BaseStatementList.debug_print(self, indentation + 2)

    def get_action_list(self):
        global item_feature, item_id, item_size
        item_id = self.id
        item_feature = self.feature.value
        item_size = self.size
        return base_statement.BaseStatementList.get_action_list(self)

    def __str__(self):
        ret = "item({:d}".format(self.feature.value)
        if self.name is not None:
            ret += ", {}".format(self.name)
        ret += ", {}".format(str(self.id) if self.id is not None else "-1")
        if self.size is not None:
            sizes = ["1X1", None, "2X1", "1X2", "2X2"]
            ret += ", HOUSE_SIZE_{}".format(sizes[self.size.value])
        ret += ") {\n"
        ret += base_statement.BaseStatementList.__str__(self)
        ret += "}\n"
        return ret


class Property:
    """
    AST-node representing a single property. These are only valid
    inside a PropertyBlock.

    @ivar name: The name (or number) of this property.
    @type name: L{Identifier} or L{ConstantNumeric}.

    @ivar value: The value that will be assigned to this property.
    @type value: L{Expression}.

    @ivar unit: The unit of the value.
    @type unit: L{Unit}

    @ivar pos: Position information.
    @type pos: L{Position}
    """

    def __init__(self, name, value, unit, pos):
        self.pos = pos
        self.name = name
        self.value = value
        self.unit = unit

    def pre_process(self):
        self.value = self.value.reduce(global_constants.const_list, unknown_id_fatal=False)

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Property:", self.name.value)
        self.value.debug_print(indentation + 2)

    def __str__(self):
        unit = "" if self.unit is None else " " + str(self.unit)
        return "\t{}: {}{};".format(self.name, self.value, unit)


class PropertyBlock(base_statement.BaseStatement):
    """
    Block that contains a list of property/value pairs to be assigned
    to the current item.

    @ivar prop_list: List of properties.
    @type prop_list: C{list} of L{Property}
    """

    def __init__(self, prop_list, pos):
        base_statement.BaseStatement.__init__(self, "property-block", pos, in_item=True, out_item=False)
        self.prop_list = prop_list

    def pre_process(self):
        for prop in self.prop_list:
            prop.pre_process()

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Property block:")
        for prop in self.prop_list:
            prop.debug_print(indentation + 2)

    def get_action_list(self):
        return action0.parse_property_block(self.prop_list, item_feature, item_id, item_size)

    def __str__(self):
        ret = "property {\n"
        for prop in self.prop_list:
            ret += "{}\n".format(prop)
        ret += "}\n"
        return ret


class LiveryOverride(base_statement.BaseStatement):
    def __init__(self, wagon_id, graphics_block, pos):
        base_statement.BaseStatement.__init__(self, "livery override", pos, in_item=True, out_item=False)
        self.graphics_block = graphics_block
        self.wagon_id = wagon_id

    def pre_process(self):
        self.graphics_block.pre_process()

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Livery override, wagon id:")
        self.wagon_id.debug_print(indentation + 2)
        for graphics in self.graphics_block.graphics_list:
            graphics.debug_print(indentation + 2)
        generic.print_dbg(indentation + 2, "Default graphics:", self.graphics_block.default_graphics)

    def get_action_list(self):
        wagon_id = self.wagon_id.reduce_constant([(global_constants.item_names, global_constants.item_to_id)])
        return action3.parse_graphics_block(self.graphics_block, item_feature, wagon_id, item_size, True)

    def __str__(self):
        ret = "livery_override({}) {{\n".format(self.wagon_id)
        for graphics in self.graphics_block.graphics_list:
            ret += "\t{}\n".format(graphics)
        if self.graphics_block.default_graphics is not None:
            ret += "\t{}\n".format(self.graphics_block.default_graphics)
        ret += "}\n"
        return ret


graphics_base_class = action2.make_sprite_group_class(False, False, True)


class GraphicsBlock(graphics_base_class):
    def __init__(self, graphics_list, default_graphics, pos):
        base_statement.BaseStatement.__init__(self, "graphics-block", pos, in_item=True, out_item=False)
        self.graphics_list = graphics_list
        self.default_graphics = default_graphics

    def pre_process(self):
        for graphics_def in self.graphics_list:
            graphics_def.reduce_expressions(action2var.get_scope(item_feature))
        if self.default_graphics is not None:
            if self.default_graphics.value is None:
                raise generic.ScriptError(
                    "Returning the computed value is not possible in a graphics-block, as there is no computed value.",
                    self.pos,
                )
            self.default_graphics.value = action2var.reduce_varaction2_expr(
                self.default_graphics.value, action2var.get_scope(item_feature)
            )

        # initialize base class and pre_process it as well (in that order)
        self.initialize(None, item_feature)
        graphics_base_class.pre_process(self)

    def collect_references(self):
        all_refs = []
        for result in [g.result for g in self.graphics_list] + (
            [self.default_graphics] if self.default_graphics is not None else []
        ):
            all_refs += result.value.collect_references()
        return all_refs

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Graphics block:")
        for graphics in self.graphics_list:
            graphics.debug_print(indentation + 2)
        if self.default_graphics is not None:
            generic.print_dbg(indentation + 2, "Default graphics:")
            self.default_graphics.debug_print(indentation + 4)

    def get_action_list(self):
        if self.prepare_act2_output():
            return action3.parse_graphics_block(self, item_feature, item_id, item_size)
        return []

    def __str__(self):
        ret = "graphics {\n"
        for graphics in self.graphics_list:
            ret += "\t{}\n".format(graphics)
        if self.default_graphics is not None:
            ret += "\t{}\n".format(self.default_graphics)
        ret += "}\n"
        return ret


class GraphicsDefinition:
    def __init__(self, cargo_id, result, unit=None):
        self.cargo_id = cargo_id
        self.result = result
        self.unit = unit

    def reduce_expressions(self, var_scope):
        # Do not reduce cargo-id (yet)
        if self.result.value is None:
            raise generic.ScriptError(
                "Returning the computed value is not possible in a graphics-block, as there is no computed value.",
                self.result.pos,
            )
        self.result.value = action2var.reduce_varaction2_expr(self.result.value, var_scope)

    def debug_print(self, indentation):
        generic.print_dbg(indentation, "Cargo ID:")
        self.cargo_id.debug_print(indentation + 2)
        generic.print_dbg(indentation, "Result:")
        self.result.debug_print(indentation + 2)

    def __str__(self):
        return "{}: {}".format(self.cargo_id, self.result)



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

from nml import expression, generic
from nml.actions import base_action, real_sprite


class Action12(base_action.BaseAction):
    # sets: list of (font_size, num_char, base_char) tuples
    def __init__(self, sets):
        self.sets = sets

    def write(self, file):
        # <sprite-number> * <length> 12 <num-def> (<font> <num-char> <base-char>){n}
        size = 2 + 4 * len(self.sets)
        file.start_sprite(size)
        file.print_bytex(0x12)
        file.print_byte(len(self.sets))
        file.newline()
        for font_size, num_char, base_char in self.sets:
            font_size.write(file, 1)
            file.print_byte(num_char)
            file.print_word(base_char)
            file.newline()
        file.end_sprite()


font_sizes = {
    "NORMAL": 0,
    "SMALL": 1,
    "LARGE": 2,
    "MONO": 3,
}


def parse_action12(font_glyphs):
    try:
        font_size = font_glyphs.font_size.reduce_constant([font_sizes])
        if isinstance(font_glyphs.base_char, expression.StringLiteral) and len(font_glyphs.base_char.value) == 1:
            base_char = ord(font_glyphs.base_char.value)
        else:
            base_char = font_glyphs.base_char.reduce_constant()
    except generic.ConstError:
        raise generic.ScriptError("Parameters of font_glyph have to be compile-time constants", font_glyphs.pos)
    if font_size.value not in font_sizes.values():
        raise generic.ScriptError(
            "Invalid value for parameter 'font_size' in font_glyph, valid values are 0, 1, 2", font_size.pos
        )
    if not (0 <= base_char.value <= 0xFFFF):
        raise generic.ScriptError(
            "Invalid value for parameter 'base_char' in font_glyph, valid values are 0-0xFFFF", base_char.pos
        )

    real_sprite_list = real_sprite.parse_sprite_data(font_glyphs)
    char = base_char.value
    last_char = char + len(real_sprite_list)
    if last_char > 0xFFFF:
        raise generic.ScriptError(
            "Character numbers in font_glyph block exceed the allowed range (0-0xFFFF)", font_glyphs.pos
        )

    sets = []
    while char < last_char:
        # each set of characters must fit in a single 128-char block, according to specs / TTDP
        if (char // 128) * 128 != (last_char // 128) * 128:
            num_in_set = (char // 128 + 1) * 128 - char
        else:
            num_in_set = last_char - char
        sets.append((font_size, num_in_set, char))
        char += num_in_set

    return [Action12(sets)] + real_sprite_list



__license__ = """
NML is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

NML is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with NML; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA."""

import operator

from nml import generic

from .expression import binop
from .expression.base_expression import ConstantFloat, ConstantNumeric, Type


class Operator:
    """
    Operator in an expression.

    @ivar act2_supports: Whether the operator is supported by action 2.
    @type act2_supports: C{bool}

    @ivar act2_str: NFO text representation of the operator in action 2.
    @type act2_str: C{str} or C{None}

    @ivar act2_num: Numeric value of the operator in action 2.
    @type act2_num: C{int} or C{None}

    @ivar actd_supports: Whether the operator is supported by action D.
    @type actd_supports: C{bool}

    @ivar actd_str: NFO text representation of the operator in action D.
    @type actd_str: C{str} or C{None}

    @ivar actd_num: Numeric value of the operator in action D.
    @type actd_num: C{int} or C{None}

    @ivar commutative: Whether swapping the operands will always maintain the same result.
    @type commutative: C{Bool}

    @ivar returns_boolean: Whether the operator gives a boolean result.
    @type returns_boolean: C{bool}

    @ivar token: Infix text to use for the string representation of the operator.
    @type token: C{None} or C{str}

    @ivar prefix_text: Prefix text to use for the string representation of the operator, if available.
    @type prefix_text: C{None} or C{str}
    """

    def __init__(
        self,
        act2_supports=False,
        act2_str=None,
        act2_num=None,
        actd_supports=False,
        actd_str=None,
        actd_num=None,
        returns_boolean=False,
        commutative=False,
        token=None,
        prefix_text=None,
        compiletime_func=None,
        validate_func=None,
    ):
        self.act2_supports = act2_supports
        self.act2_str = act2_str
        self.act2_num = act2_num
        self.actd_supports = actd_supports
        self.actd_str = actd_str
        self.actd_num = actd_num
        self.commutative = commutative
        self.returns_boolean = returns_boolean
        self.token = token
        self.prefix_text = prefix_text
        self.compiletime_func = compiletime_func
        self.validate_func = validate_func

    def to_string(self, expr1, expr2):
        """
        Convert expression to readable string form.

        @param expr1: Left child expression text.
        @type  expr1: C{str}

        @param expr2: Right child expression text.
        @type  expr2: C{str}

        @return: Text representation of the operator with its child expressions.
        @rtype:  C{str}
        """
        if self.prefix_text is not None:
            return "{}({}, {})".format(self.prefix_text, expr1, expr2)
        else:  # Infix notation.
            return "({} {} {})".format(expr1, self.token, expr2)

    def __call__(self, expr1, expr2, pos=None):
        return binop.BinOp(self, expr1, expr2, pos)


def unsigned_rshift(a, b):
    if a < 0:
        a += 0x100000000
    return generic.truncate_int32(a >> b)


def unsigned_rrotate(a, b):
    if a < 0:
        a += 0x100000000
    return generic.truncate_int32((a >> b) | (a << (32 - b)))


def validate_func_int(expr1, expr2, pos):
    if expr1.type() != Type.INTEGER or expr2.type() != Type.INTEGER:
        if expr1.type() == Type.SPRITEGROUP_REF:
            raise generic.ProcCallSyntaxError(expr1.name, expr1.pos)
        if expr2.type() == Type.SPRITEGROUP_REF:
            raise generic.ProcCallSyntaxError(expr2.name, expr2.pos)
        raise generic.ScriptError("Binary operator requires both operands to be integers.", pos)


def validate_func_float(expr1, expr2, pos):
    if expr1.type() not in (Type.INTEGER, Type.FLOAT) or expr2.type() not in (Type.INTEGER, Type.FLOAT):
        if expr1.type() == Type.SPRITEGROUP_REF:
            raise generic.ProcCallSyntaxError(expr1.name, expr1.pos)
        if expr2.type() == Type.SPRITEGROUP_REF:
            raise generic.ProcCallSyntaxError(expr2.name, expr2.pos)
        raise generic.ScriptError("Binary operator requires both operands to be integers or floats.", pos)
    # If one is a float, the other must be constant since we can't handle floats at runtime
    if (expr1.type() == Type.FLOAT and not isinstance(expr2, (ConstantNumeric, ConstantFloat))) or (
        expr2.type() == Type.FLOAT and not isinstance(expr1, (ConstantNumeric, ConstantFloat))
    ):
        raise generic.ScriptError(
            "Floating-point operations are only possible when both operands are compile-time constants.", pos
        )


def validate_func_add(expr1, expr2, pos):
    if (expr1.type() == Type.STRING_LITERAL) ^ (expr2.type() == Type.STRING_LITERAL):
        raise generic.ScriptError("Concatenating a string literal and a number is not possible.", pos)
    if expr1.type() != Type.STRING_LITERAL:
        validate_func_float(expr1, expr2, pos)


def validate_func_div_mod(expr1, expr2, pos):
    validate_func_float(expr1, expr2, pos)
    if isinstance(expr2, (ConstantNumeric, ConstantFloat)) and expr2.value == 0:
        raise generic.ScriptError("Division and modulo require the right hand side to be nonzero.", pos)


def validate_func_rhs_positive(expr1, expr2, pos):
    validate_func_int(expr1, expr2, pos)
    if isinstance(expr2, ConstantNumeric) and expr2.value < 0:
        raise generic.ScriptError("Right hand side of the operator may not be a negative number.", pos)


ADD = Operator(
    act2_supports=True,
    act2_str=r"\2+",
    act2_num=0,
    actd_supports=True,
    actd_str=r"\D+",
    actd_num=1,
    commutative=True,
    token="+",
    compiletime_func=operator.add,
    validate_func=validate_func_add,
)

SUB = Operator(
    act2_supports=True,
    act2_str=r"\2-",
    act2_num=1,
    actd_supports=True,
    actd_str=r"\D-",
    actd_num=2,
    token="-",
    compiletime_func=operator.sub,
    validate_func=validate_func_float,
)

DIV = Operator(
    act2_supports=True,
    act2_str=r"\2/",
    act2_num=6,
    actd_supports=True,
    actd_str=r"\D/",
    actd_num=10,
    token="/",
    compiletime_func=lambda a, b: a // b if isinstance(a, int) and isinstance(b, int) else a / b,
    validate_func=validate_func_div_mod,
)

MOD = Operator(
    act2_supports=True,
    act2_str=r"\2%",
    act2_num=7,
    actd_supports=True,
    actd_str=r"\D%",
    actd_num=12,
    token="%",
    compiletime_func=operator.mod,
    validate_func=validate_func_div_mod,
)

MUL = Operator(
    act2_supports=True,
    act2_str=r"\2*",
    act2_num=10,
    actd_supports=True,
    actd_str=r"\D*",
    actd_num=4,
    commutative=True,
    token="*",
    compiletime_func=operator.mul,
    validate_func=validate_func_float,
)

AND = Operator(
    act2_supports=True,
    act2_str=r"\2&",
    act2_num=11,
    actd_supports=True,
    actd_str=r"\D&",
    actd_num=7,
    commutative=True,
    token="&",
    compiletime_func=operator.and_,
    validate_func=validate_func_int,
)

OR = Operator(
    act2_supports=True,
    act2_str=r"\2|",
    act2_num=12,
    actd_supports=True,
    actd_str=r"\D|",
    actd_num=8,
    commutative=True,
    token="|",
    compiletime_func=operator.or_,
    validate_func=validate_func_int,
)

XOR = Operator(
    act2_supports=True,
    act2_str=r"\2^",
    act2_num=13,
    actd_supports=True,
    commutative=True,
    token="^",
    compiletime_func=operator.xor,
    validate_func=validate_func_int,
)

CMP_EQ = Operator(
    act2_supports=True,
    actd_supports=True,
    commutative=True,
    returns_boolean=True,
    token="==",
    compiletime_func=operator.eq,
    validate_func=validate_func_float,
)

CMP_NEQ = Operator(
    act2_supports=True,
    actd_supports=True,
    commutative=True,
    returns_boolean=True,
    token="!=",
    compiletime_func=operator.ne,
    validate_func=validate_func_float,
)

CMP_LE = Operator(
    act2_supports=True,
    actd_supports=True,
    returns_boolean=True,
    token="<=",
    compiletime_func=operator.le,
    validate_func=validate_func_float,
)

CMP_GE = Operator(
    act2_supports=True,
    actd_supports=True,
    returns_boolean=True,
    token=">=",
    compiletime_func=operator.ge,
    validate_func=validate_func_float,
)

CMP_LT = Operator(
    act2_supports=True,
    actd_supports=True,
    returns_boolean=True,
    token="<",
    compiletime_func=operator.lt,
    validate_func=validate_func_float,
)

CMP_GT = Operator(
    act2_supports=True,
    actd_supports=True,
    returns_boolean=True,
    token=">",
    compiletime_func=operator.gt,
    validate_func=validate_func_float,
)

MIN = Operator(
    act2_supports=True,
    act2_str=r"\2<",
    act2_num=2,
    actd_supports=True,
    commutative=True,
    compiletime_func=min,
    prefix_text="min",
    validate_func=validate_func_float,
)

MAX = Operator(
    act2_supports=True,
    act2_str=r"\2>",
    act2_num=3,
    actd_supports=True,
    commutative=True,
    compiletime_func=max,
    prefix_text="max",
    validate_func=validate_func_float,
)

STO_TMP = Operator(
    act2_supports=True,
    act2_str=r"\2sto",
    act2_num=14,
    prefix_text="STORE_TEMP",
    validate_func=validate_func_rhs_positive,
)

STO_PERM = Operator(
    act2_supports=True,
    act2_str=r"\2psto",
    act2_num=16,
    prefix_text="STORE_PERM",
    validate_func=validate_func_rhs_positive,
)

SHIFT_LEFT = Operator(
    act2_supports=True,
    act2_str=r"\2<<",
    act2_num=20,
    actd_supports=True,
    actd_str=r"\D<<",
    actd_num=6,
    token="<<",
    compiletime_func=operator.lshift,
    validate_func=validate_func_rhs_positive,
)

SHIFT_RIGHT = Operator(
    act2_supports=True,
    act2_str=r"\2>>",
    act2_num=22,
    actd_supports=True,
    token=">>",
    compiletime_func=operator.rshift,
    validate_func=validate_func_rhs_positive,
)

SHIFTU_RIGHT = Operator(
    act2_supports=True,
    act2_str=r"\2u>>",
    act2_num=21,
    actd_supports=True,
    token=">>>",
    compiletime_func=unsigned_rshift,
    validate_func=validate_func_rhs_positive,
)

HASBIT = Operator(
    act2_supports=True,
    actd_supports=True,
    returns_boolean=True,
    prefix_text="hasbit",
    compiletime_func=lambda a, b: (a & (1 << b)) != 0,
    validate_func=validate_func_rhs_positive,
)

# A few operators that are generated internally but can't be directly written in nml
NOTHASBIT = Operator(
    act2_supports=True,
    actd_supports=True,
    returns_boolean=True,
    prefix_text="!hasbit",
    compiletime_func=lambda a, b: (a & (1 << b)) == 0,
)

VAL2 = Operator(
    act2_supports=True,
    act2_str=r"\2r",
    act2_num=15,
    compiletime_func=lambda a, b: b,
)

ASSIGN = Operator(
    actd_supports=True,
    actd_str=r"\D=",
    actd_num=0,
)

SHIFTU_LEFT = Operator(
    actd_supports=True,
    actd_str=r"\Du<<",
    actd_num=5,
    token="<<",
)

VACT2_CMP = Operator(
    act2_supports=True,
    act2_str=r"\2cmp",
    act2_num=18,
    prefix_text="CMP",
)

VACT2_UCMP = Operator(
    act2_supports=True,
    act2_str=r"\2ucmp",
    act2_num=19,
    prefix_text="UCMP",
)

MINU = Operator(
    act2_supports=True,
    act2_str=r"\2u<",
    act2_num=4,
)

ROT_RIGHT = Operator(
    act2_supports=True,
    act2_str=r"\2ror",
    act2_num=17,
    prefix_text="rotate",
    compiletime_func=unsigned_rrotate,
    validate_func=validate_func_int,
)

DIVU = Operator(
    act2_supports=True,
    act2_str=r"\2u/",
    act2_num=8,
    actd_supports=True,
    actd_str=r"\Du/",
    actd_num=9,
)


class GRMOperator:
    def __init__(self, op_str, op_num):
        self.op_str = op_str
        self.op_num = op_num
        self.value = op_num

    def __str__(self):
        return self.op_str

    def write(self, file, size):
        assert size == 1
        file.print_bytex(self.op_num, self.op_str)


GRM_RESERVE = GRMOperator(r"\DR", 0)
